# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..brainsfit import BRAINSFit


def test_BRAINSFit_inputs():
    input_map = dict(
        NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_00=dict(
            argstr="--NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_00 ",
        ),
        NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_01=dict(
            argstr="--NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_01 ",
        ),
        NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_02=dict(
            argstr="--NEVER_USE_THIS_FLAG_IT_IS_OUTDATED_02 ",
        ),
        ROIAutoClosingSize=dict(
            argstr="--ROIAutoClosingSize %f",
        ),
        ROIAutoDilateSize=dict(
            argstr="--ROIAutoDilateSize %f",
        ),
        args=dict(
            argstr="%s",
        ),
        backgroundFillValue=dict(
            argstr="--backgroundFillValue %f",
        ),
        bsplineTransform=dict(
            argstr="--bsplineTransform %s",
            hash_files=False,
        ),
        costFunctionConvergenceFactor=dict(
            argstr="--costFunctionConvergenceFactor %f",
        ),
        costMetric=dict(
            argstr="--costMetric %s",
        ),
        debugLevel=dict(
            argstr="--debugLevel %d",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        failureExitCode=dict(
            argstr="--failureExitCode %d",
        ),
        fixedBinaryVolume=dict(
            argstr="--fixedBinaryVolume %s",
            extensions=None,
        ),
        fixedVolume=dict(
            argstr="--fixedVolume %s",
            extensions=None,
        ),
        fixedVolumeTimeIndex=dict(
            argstr="--fixedVolumeTimeIndex %d",
        ),
        forceMINumberOfThreads=dict(
            argstr="--forceMINumberOfThreads %d",
        ),
        gui=dict(
            argstr="--gui ",
        ),
        histogramMatch=dict(
            argstr="--histogramMatch ",
        ),
        initialTransform=dict(
            argstr="--initialTransform %s",
            extensions=None,
        ),
        initializeTransformMode=dict(
            argstr="--initializeTransformMode %s",
        ),
        interpolationMode=dict(
            argstr="--interpolationMode %s",
        ),
        linearTransform=dict(
            argstr="--linearTransform %s",
            hash_files=False,
        ),
        maskInferiorCutOffFromCenter=dict(
            argstr="--maskInferiorCutOffFromCenter %f",
        ),
        maskProcessingMode=dict(
            argstr="--maskProcessingMode %s",
        ),
        maxBSplineDisplacement=dict(
            argstr="--maxBSplineDisplacement %f",
        ),
        maximumStepLength=dict(
            argstr="--maximumStepLength %f",
        ),
        medianFilterSize=dict(
            argstr="--medianFilterSize %s",
            sep=",",
        ),
        minimumStepLength=dict(
            argstr="--minimumStepLength %s",
            sep=",",
        ),
        movingBinaryVolume=dict(
            argstr="--movingBinaryVolume %s",
            extensions=None,
        ),
        movingVolume=dict(
            argstr="--movingVolume %s",
            extensions=None,
        ),
        movingVolumeTimeIndex=dict(
            argstr="--movingVolumeTimeIndex %d",
        ),
        numberOfHistogramBins=dict(
            argstr="--numberOfHistogramBins %d",
        ),
        numberOfIterations=dict(
            argstr="--numberOfIterations %s",
            sep=",",
        ),
        numberOfMatchPoints=dict(
            argstr="--numberOfMatchPoints %d",
        ),
        numberOfSamples=dict(
            argstr="--numberOfSamples %d",
        ),
        numberOfThreads=dict(
            argstr="--numberOfThreads %d",
        ),
        outputFixedVolumeROI=dict(
            argstr="--outputFixedVolumeROI %s",
            hash_files=False,
        ),
        outputMovingVolumeROI=dict(
            argstr="--outputMovingVolumeROI %s",
            hash_files=False,
        ),
        outputTransform=dict(
            argstr="--outputTransform %s",
            hash_files=False,
        ),
        outputVolume=dict(
            argstr="--outputVolume %s",
            hash_files=False,
        ),
        outputVolumePixelType=dict(
            argstr="--outputVolumePixelType %s",
        ),
        permitParameterVariation=dict(
            argstr="--permitParameterVariation %s",
            sep=",",
        ),
        projectedGradientTolerance=dict(
            argstr="--projectedGradientTolerance %f",
        ),
        promptUser=dict(
            argstr="--promptUser ",
        ),
        relaxationFactor=dict(
            argstr="--relaxationFactor %f",
        ),
        removeIntensityOutliers=dict(
            argstr="--removeIntensityOutliers %f",
        ),
        reproportionScale=dict(
            argstr="--reproportionScale %f",
        ),
        scaleOutputValues=dict(
            argstr="--scaleOutputValues ",
        ),
        skewScale=dict(
            argstr="--skewScale %f",
        ),
        splineGridSize=dict(
            argstr="--splineGridSize %s",
            sep=",",
        ),
        strippedOutputTransform=dict(
            argstr="--strippedOutputTransform %s",
            hash_files=False,
        ),
        transformType=dict(
            argstr="--transformType %s",
            sep=",",
        ),
        translationScale=dict(
            argstr="--translationScale %f",
        ),
        useAffine=dict(
            argstr="--useAffine ",
        ),
        useBSpline=dict(
            argstr="--useBSpline ",
        ),
        useCachingOfBSplineWeightsMode=dict(
            argstr="--useCachingOfBSplineWeightsMode %s",
        ),
        useExplicitPDFDerivativesMode=dict(
            argstr="--useExplicitPDFDerivativesMode %s",
        ),
        useRigid=dict(
            argstr="--useRigid ",
        ),
        useScaleSkewVersor3D=dict(
            argstr="--useScaleSkewVersor3D ",
        ),
        useScaleVersor3D=dict(
            argstr="--useScaleVersor3D ",
        ),
        writeOutputTransformInFloat=dict(
            argstr="--writeOutputTransformInFloat ",
        ),
        writeTransformOnFailure=dict(
            argstr="--writeTransformOnFailure ",
        ),
    )
    inputs = BRAINSFit.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_BRAINSFit_outputs():
    output_map = dict(
        bsplineTransform=dict(
            extensions=None,
        ),
        linearTransform=dict(
            extensions=None,
        ),
        outputFixedVolumeROI=dict(
            extensions=None,
        ),
        outputMovingVolumeROI=dict(
            extensions=None,
        ),
        outputTransform=dict(
            extensions=None,
        ),
        outputVolume=dict(
            extensions=None,
        ),
        strippedOutputTransform=dict(
            extensions=None,
        ),
    )
    outputs = BRAINSFit.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



from nipype.pipeline.plugins.base import SGELikeBatchManagerBase
from nipype.interfaces.utility import Function
import nipype.pipeline.engine as pe
import pytest
from unittest.mock import patch
import subprocess


def crasher():
    raise ValueError()


def submit_batchtask(self, scriptfile, node):
    self._pending[1] = node.output_dir()
    subprocess.call(["bash", scriptfile])
    return 1


def is_pending(self, taskid):
    return False


@patch.object(SGELikeBatchManagerBase, "_submit_batchtask", new=submit_batchtask)
@patch.object(SGELikeBatchManagerBase, "_is_pending", new=is_pending)
def test_crashfile_creation(tmp_path):
    pipe = pe.Workflow(name="pipe", base_dir=str(tmp_path))
    pipe.config["execution"]["crashdump_dir"] = str(tmp_path)
    pipe.add_nodes([pe.Node(interface=Function(function=crasher), name="crasher")])
    sgelike_plugin = SGELikeBatchManagerBase("")
    with pytest.raises(RuntimeError) as e:
        assert pipe.run(plugin=sgelike_plugin)

    crashfiles = list(tmp_path.glob("crash*crasher*.pklz")) + list(
        tmp_path.glob("crash*crasher*.txt")
    )
    assert len(crashfiles) == 1



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utilities import GenerateEdgeMapImage


def test_GenerateEdgeMapImage_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputMRVolumes=dict(
            argstr="--inputMRVolumes %s...",
        ),
        inputMask=dict(
            argstr="--inputMask %s",
            extensions=None,
        ),
        lowerPercentileMatching=dict(
            argstr="--lowerPercentileMatching %f",
        ),
        maximumOutputRange=dict(
            argstr="--maximumOutputRange %d",
        ),
        minimumOutputRange=dict(
            argstr="--minimumOutputRange %d",
        ),
        numberOfThreads=dict(
            argstr="--numberOfThreads %d",
        ),
        outputEdgeMap=dict(
            argstr="--outputEdgeMap %s",
            hash_files=False,
        ),
        outputMaximumGradientImage=dict(
            argstr="--outputMaximumGradientImage %s",
            hash_files=False,
        ),
        upperPercentileMatching=dict(
            argstr="--upperPercentileMatching %f",
        ),
    )
    inputs = GenerateEdgeMapImage.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_GenerateEdgeMapImage_outputs():
    output_map = dict(
        outputEdgeMap=dict(
            extensions=None,
        ),
        outputMaximumGradientImage=dict(
            extensions=None,
        ),
    )
    outputs = GenerateEdgeMapImage.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import TShift


def test_TShift_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore=dict(
            argstr="-ignore %s",
        ),
        in_file=dict(
            argstr="%s",
            copyfile=False,
            extensions=None,
            mandatory=True,
            position=-1,
        ),
        interp=dict(
            argstr="-%s",
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_file=dict(
            argstr="-prefix %s",
            extensions=None,
            name_source="in_file",
            name_template="%s_tshift",
        ),
        outputtype=dict(),
        rlt=dict(
            argstr="-rlt",
        ),
        rltplus=dict(
            argstr="-rlt+",
        ),
        slice_encoding_direction=dict(
            usedefault=True,
        ),
        slice_timing=dict(
            argstr="-tpattern @%s",
            xor=["tpattern"],
        ),
        tpattern=dict(
            argstr="-tpattern %s",
            xor=["slice_timing"],
        ),
        tr=dict(
            argstr="-TR %s",
        ),
        tslice=dict(
            argstr="-slice %s",
            xor=["tzero"],
        ),
        tzero=dict(
            argstr="-tzero %s",
            xor=["tslice"],
        ),
    )
    inputs = TShift.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_TShift_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
        timing_file=dict(
            extensions=None,
        ),
    )
    outputs = TShift.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..base import FSCommand


def test_FSCommand_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        subjects_dir=dict(),
    )
    inputs = FSCommand.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import CAT12SANLMDenoising


def test_CAT12SANLMDenoising_inputs():
    input_map = dict(
        addnoise=dict(
            field="addnoise",
            usedefault=True,
        ),
        filename_prefix=dict(
            field="prefix",
            usedefault=True,
        ),
        filename_suffix=dict(
            field="suffix",
            usedefault=True,
        ),
        in_files=dict(
            copyfile=False,
            field="data",
            mandatory=True,
        ),
        intlim=dict(
            field="intlim",
            usedefault=True,
        ),
        matlab_cmd=dict(),
        mfile=dict(
            usedefault=True,
        ),
        noisecorr_strength=dict(
            field="nlmfilter.optimized.NCstr",
            usedefault=True,
        ),
        paths=dict(),
        replace_nan_and_inf=dict(
            field="replaceNANandINF",
            usedefault=True,
        ),
        rician=dict(
            field="rician",
            usedefault=True,
        ),
        spm_type=dict(
            field="spm_type",
            usedefault=True,
        ),
        use_mcr=dict(),
        use_v8struct=dict(
            min_ver="8",
            usedefault=True,
        ),
    )
    inputs = CAT12SANLMDenoising.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_CAT12SANLMDenoising_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = CAT12SANLMDenoising.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



"""ANTs' utilities."""
import os
from warnings import warn
from ..base import traits, isdefined, TraitedSpec, File, Str, InputMultiObject
from ..mixins import CopyHeaderInterface
from .base import ANTSCommandInputSpec, ANTSCommand


class ImageMathInputSpec(ANTSCommandInputSpec):
    dimension = traits.Int(
        3, usedefault=True, position=1, argstr="%d", desc="dimension of output image"
    )
    output_image = File(
        position=2,
        argstr="%s",
        name_source=["op1"],
        name_template="%s_maths",
        desc="output image file",
        keep_extension=True,
    )
    operation = traits.Enum(
        # Mathematical Operations
        "m",
        "vm",
        "+",
        "v+",
        "-",
        "v-",
        "/",
        "^",
        "max",
        "exp",
        "addtozero",
        "overadd",
        "abs",
        "total",
        "mean",
        "vtotal",
        "Decision",
        "Neg",
        # Spatial Filtering Operations
        "Project",
        "G",
        "MD",
        "ME",
        "MO",
        "MC",
        "GD",
        "GE",
        "GO",
        "GC",
        "ExtractContours",
        # Transform Image Operations
        "Translate",
        # Tensor Operations
        "4DTensorTo3DTensor",
        "ExtractVectorComponent",
        "TensorColor",
        "TensorFA",
        "TensorFADenominator",
        "TensorFANumerator",
        "TensorMeanDiffusion",
        "TensorRadialDiffusion",
        "TensorAxialDiffusion",
        "TensorEigenvalue",
        "TensorToVector",
        "TensorToVectorComponent",
        "TensorMask",
        # Unclassified Operators
        "Byte",
        "CorruptImage",
        "D",
        "MaurerDistance",
        "ExtractSlice",
        "FillHoles",
        "Convolve",
        "Finite",
        "FlattenImage",
        "GetLargestComponent",
        "Grad",
        "RescaleImage",
        "WindowImage",
        "NeighborhoodStats",
        "ReplicateDisplacement",
        "ReplicateImage",
        "LabelStats",
        "Laplacian",
        "Canny",
        "Lipschitz",
        "MTR",
        "Normalize",
        "PadImage",
        "SigmoidImage",
        "Sharpen",
        "UnsharpMask",
        "PValueImage",
        "ReplaceVoxelValue",
        "SetTimeSpacing",
        "SetTimeSpacingWarp",
        "stack",
        "ThresholdAtMean",
        "TriPlanarView",
        "TruncateImageIntensity",
        mandatory=True,
        position=3,
        argstr="%s",
        desc="mathematical operations",
    )
    op1 = File(
        exists=True, mandatory=True, position=-3, argstr="%s", desc="first operator"
    )
    op2 = traits.Either(
        File(exists=True), Str, position=-2, argstr="%s", desc="second operator"
    )

    args = Str(position=-1, argstr="%s", desc="Additional parameters to the command")

    copy_header = traits.Bool(
        True,
        usedefault=True,
        desc="copy headers of the original image into the output (corrected) file",
    )


class ImageMathOuputSpec(TraitedSpec):
    output_image = File(exists=True, desc="output image file")


class ImageMath(ANTSCommand, CopyHeaderInterface):
    """
    Operations over images.

    Examples
    --------
    >>> ImageMath(
    ...     op1='structural.nii',
    ...     operation='+',
    ...     op2='2').cmdline
    'ImageMath 3 structural_maths.nii + structural.nii 2'

    >>> ImageMath(
    ...     op1='structural.nii',
    ...     operation='Project',
    ...     op2='1 2').cmdline
    'ImageMath 3 structural_maths.nii Project structural.nii 1 2'

    >>> ImageMath(
    ...     op1='structural.nii',
    ...     operation='G',
    ...     op2='4').cmdline
    'ImageMath 3 structural_maths.nii G structural.nii 4'

    >>> ImageMath(
    ...     op1='structural.nii',
    ...     operation='TruncateImageIntensity',
    ...     op2='0.005 0.999 256').cmdline
    'ImageMath 3 structural_maths.nii TruncateImageIntensity structural.nii 0.005 0.999 256'

    By default, Nipype copies headers from the first input image (``op1``)
    to the output image.
    For some operations, as the ``PadImage`` operation, the header cannot be copied from inputs to
    outputs, and so ``copy_header`` option is automatically set to ``False``.

    >>> pad = ImageMath(
    ...     op1='structural.nii',
    ...     operation='PadImage')
    >>> pad.inputs.copy_header
    False

    While the operation is set to ``PadImage``,
    setting ``copy_header = True`` will have no effect.

    >>> pad.inputs.copy_header = True
    >>> pad.inputs.copy_header
    False

    For any other operation, ``copy_header`` can be enabled/disabled normally:

    >>> pad.inputs.operation = "ME"
    >>> pad.inputs.copy_header = True
    >>> pad.inputs.copy_header
    True

    """

    _cmd = "ImageMath"
    input_spec = ImageMathInputSpec
    output_spec = ImageMathOuputSpec
    _copy_header_map = {"output_image": "op1"}
    _no_copy_header_operation = (
        "PadImage",
        "LabelStats",
        "SetTimeSpacing",
        "SetTimeSpacingWarp",
        "TriPlanarView",
    )

    def __init__(self, **inputs):
        super(ImageMath, self).__init__(**inputs)
        if self.inputs.operation in self._no_copy_header_operation:
            self.inputs.copy_header = False

        self.inputs.on_trait_change(self._operation_update, "operation")
        self.inputs.on_trait_change(self._copyheader_update, "copy_header")

    def _operation_update(self):
        if self.inputs.operation in self._no_copy_header_operation:
            self.inputs.copy_header = False

    def _copyheader_update(self):
        if (
            self.inputs.copy_header
            and self.inputs.operation in self._no_copy_header_operation
        ):
            warn(
                f"copy_header cannot be updated to True with {self.inputs.operation} as operation."
            )
            self.inputs.copy_header = False


class ResampleImageBySpacingInputSpec(ANTSCommandInputSpec):
    dimension = traits.Int(
        3, usedefault=True, position=1, argstr="%d", desc="dimension of output image"
    )
    input_image = File(
        exists=True, mandatory=True, position=2, argstr="%s", desc="input image file"
    )
    output_image = File(
        position=3,
        argstr="%s",
        name_source=["input_image"],
        name_template="%s_resampled",
        desc="output image file",
        keep_extension=True,
    )
    out_spacing = traits.Either(
        traits.List(traits.Float, minlen=2, maxlen=3),
        traits.Tuple(traits.Float, traits.Float, traits.Float),
        traits.Tuple(traits.Float, traits.Float),
        position=4,
        argstr="%s",
        mandatory=True,
        desc="output spacing",
    )
    apply_smoothing = traits.Bool(
        False, argstr="%d", position=5, desc="smooth before resampling"
    )
    addvox = traits.Int(
        argstr="%d",
        position=6,
        requires=["apply_smoothing"],
        desc="addvox pads each dimension by addvox",
    )
    nn_interp = traits.Bool(
        argstr="%d", desc="nn interpolation", position=-1, requires=["addvox"]
    )


class ResampleImageBySpacingOutputSpec(TraitedSpec):
    output_image = File(exists=True, desc="resampled file")


class ResampleImageBySpacing(ANTSCommand):
    """
    Resample an image with a given spacing.

    Examples
    --------
    >>> res = ResampleImageBySpacing(dimension=3)
    >>> res.inputs.input_image = 'structural.nii'
    >>> res.inputs.output_image = 'output.nii.gz'
    >>> res.inputs.out_spacing = (4, 4, 4)
    >>> res.cmdline  #doctest: +ELLIPSIS
    'ResampleImageBySpacing 3 structural.nii output.nii.gz 4 4 4'

    >>> res = ResampleImageBySpacing(dimension=3)
    >>> res.inputs.input_image = 'structural.nii'
    >>> res.inputs.output_image = 'output.nii.gz'
    >>> res.inputs.out_spacing = (4, 4, 4)
    >>> res.inputs.apply_smoothing = True
    >>> res.cmdline  #doctest: +ELLIPSIS
    'ResampleImageBySpacing 3 structural.nii output.nii.gz 4 4 4 1'

    >>> res = ResampleImageBySpacing(dimension=3)
    >>> res.inputs.input_image = 'structural.nii'
    >>> res.inputs.output_image = 'output.nii.gz'
    >>> res.inputs.out_spacing = (0.4, 0.4, 0.4)
    >>> res.inputs.apply_smoothing = True
    >>> res.inputs.addvox = 2
    >>> res.inputs.nn_interp = False
    >>> res.cmdline  #doctest: +ELLIPSIS
    'ResampleImageBySpacing 3 structural.nii output.nii.gz 0.4 0.4 0.4 1 2 0'

    """

    _cmd = "ResampleImageBySpacing"
    input_spec = ResampleImageBySpacingInputSpec
    output_spec = ResampleImageBySpacingOutputSpec

    def _format_arg(self, name, trait_spec, value):
        if name == "out_spacing":
            if len(value) != self.inputs.dimension:
                raise ValueError("out_spacing dimensions should match dimension")

            value = " ".join(["%g" % d for d in value])

        return super(ResampleImageBySpacing, self)._format_arg(name, trait_spec, value)


class ThresholdImageInputSpec(ANTSCommandInputSpec):
    dimension = traits.Int(
        3, usedefault=True, position=1, argstr="%d", desc="dimension of output image"
    )
    input_image = File(
        exists=True, mandatory=True, position=2, argstr="%s", desc="input image file"
    )
    output_image = File(
        position=3,
        argstr="%s",
        name_source=["input_image"],
        name_template="%s_resampled",
        desc="output image file",
        keep_extension=True,
    )

    mode = traits.Enum(
        "Otsu",
        "Kmeans",
        argstr="%s",
        position=4,
        requires=["num_thresholds"],
        xor=["th_low", "th_high"],
        desc="whether to run Otsu / Kmeans thresholding",
    )
    num_thresholds = traits.Int(position=5, argstr="%d", desc="number of thresholds")
    input_mask = File(
        exists=True,
        requires=["num_thresholds"],
        argstr="%s",
        desc="input mask for Otsu, Kmeans",
    )

    th_low = traits.Float(position=4, argstr="%f", xor=["mode"], desc="lower threshold")
    th_high = traits.Float(
        position=5, argstr="%f", xor=["mode"], desc="upper threshold"
    )
    inside_value = traits.Float(
        1, position=6, argstr="%f", requires=["th_low"], desc="inside value"
    )
    outside_value = traits.Float(
        0, position=7, argstr="%f", requires=["th_low"], desc="outside value"
    )
    copy_header = traits.Bool(
        True,
        mandatory=True,
        usedefault=True,
        desc="copy headers of the original image into the output (corrected) file",
    )


class ThresholdImageOutputSpec(TraitedSpec):
    output_image = File(exists=True, desc="resampled file")


class ThresholdImage(ANTSCommand, CopyHeaderInterface):
    """
    Apply thresholds on images.

    Examples
    --------
    >>> thres = ThresholdImage(dimension=3)
    >>> thres.inputs.input_image = 'structural.nii'
    >>> thres.inputs.output_image = 'output.nii.gz'
    >>> thres.inputs.th_low = 0.5
    >>> thres.inputs.th_high = 1.0
    >>> thres.inputs.inside_value = 1.0
    >>> thres.inputs.outside_value = 0.0
    >>> thres.cmdline  #doctest: +ELLIPSIS
    'ThresholdImage 3 structural.nii output.nii.gz 0.500000 1.000000 1.000000 0.000000'

    >>> thres = ThresholdImage(dimension=3)
    >>> thres.inputs.input_image = 'structural.nii'
    >>> thres.inputs.output_image = 'output.nii.gz'
    >>> thres.inputs.mode = 'Kmeans'
    >>> thres.inputs.num_thresholds = 4
    >>> thres.cmdline  #doctest: +ELLIPSIS
    'ThresholdImage 3 structural.nii output.nii.gz Kmeans 4'

    """

    _cmd = "ThresholdImage"
    input_spec = ThresholdImageInputSpec
    output_spec = ThresholdImageOutputSpec
    _copy_header_map = {"output_image": "input_image"}


class AIInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, usedefault=True, argstr="-d %d", desc="dimension of output image"
    )
    verbose = traits.Bool(
        False, usedefault=True, argstr="-v %d", desc="enable verbosity"
    )

    fixed_image = File(
        exists=True,
        mandatory=True,
        desc="Image to which the moving_image should be transformed",
    )
    moving_image = File(
        exists=True,
        mandatory=True,
        desc="Image that will be transformed to fixed_image",
    )

    fixed_image_mask = File(exists=True, argstr="-x %s", desc="fixed mage mask")
    moving_image_mask = File(
        exists=True, requires=["fixed_image_mask"], desc="moving mage mask"
    )

    metric_trait = (
        traits.Enum("Mattes", "GC", "MI"),
        traits.Int(32),
        traits.Enum("Regular", "Random", "None"),
        traits.Range(value=0.2, low=0.0, high=1.0),
    )
    metric = traits.Tuple(
        *metric_trait, argstr="-m %s", mandatory=True, desc="the metric(s) to use."
    )

    transform = traits.Tuple(
        traits.Enum("Affine", "Rigid", "Similarity"),
        traits.Range(value=0.1, low=0.0, exclude_low=True),
        argstr="-t %s[%g]",
        usedefault=True,
        desc="Several transform options are available",
    )

    principal_axes = traits.Bool(
        False,
        usedefault=True,
        argstr="-p %d",
        xor=["blobs"],
        desc="align using principal axes",
    )
    search_factor = traits.Tuple(
        traits.Float(20),
        traits.Range(value=0.12, low=0.0, high=1.0),
        usedefault=True,
        argstr="-s [%g,%g]",
        desc="search factor",
    )

    search_grid = traits.Either(
        traits.Tuple(
            traits.Float, traits.Tuple(traits.Float, traits.Float, traits.Float)
        ),
        traits.Tuple(traits.Float, traits.Tuple(traits.Float, traits.Float)),
        argstr="-g %s",
        desc="Translation search grid in mm",
        min_ver="2.3.0",
    )

    convergence = traits.Tuple(
        traits.Range(low=1, high=10000, value=10),
        traits.Float(1e-6),
        traits.Range(low=1, high=100, value=10),
        usedefault=True,
        argstr="-c [%d,%g,%d]",
        desc="convergence",
    )

    output_transform = File(
        "initialization.mat", usedefault=True, argstr="-o %s", desc="output file name"
    )


class AIOuputSpec(TraitedSpec):
    output_transform = File(exists=True, desc="output file name")


class AI(ANTSCommand):
    """
    Calculate the optimal linear transform parameters for aligning two images.

    Examples
    --------
    >>> AI(
    ...     fixed_image='structural.nii',
    ...     moving_image='epi.nii',
    ...     metric=('Mattes', 32, 'Regular', 1),
    ... ).cmdline
    'antsAI -c [10,1e-06,10] -d 3 -m Mattes[structural.nii,epi.nii,32,Regular,1]
    -o initialization.mat -p 0 -s [20,0.12] -t Affine[0.1] -v 0'

    >>> AI(fixed_image='structural.nii',
    ...    moving_image='epi.nii',
    ...    metric=('Mattes', 32, 'Regular', 1),
    ...    search_grid=(12, (1, 1, 1)),
    ... ).cmdline
    'antsAI -c [10,1e-06,10] -d 3 -m Mattes[structural.nii,epi.nii,32,Regular,1]
    -o initialization.mat -p 0 -s [20,0.12] -g [12.0,1x1x1] -t Affine[0.1] -v 0'

    """

    _cmd = "antsAI"
    input_spec = AIInputSpec
    output_spec = AIOuputSpec

    def _run_interface(self, runtime, correct_return_codes=(0,)):
        runtime = super(AI, self)._run_interface(runtime, correct_return_codes)

        self._output = {
            "output_transform": os.path.join(
                runtime.cwd, os.path.basename(self.inputs.output_transform)
            )
        }
        return runtime

    def _format_arg(self, opt, spec, val):
        if opt == "metric":
            val = "%s[{fixed_image},{moving_image},%d,%s,%g]" % val
            val = val.format(
                fixed_image=self.inputs.fixed_image,
                moving_image=self.inputs.moving_image,
            )
            return spec.argstr % val

        if opt == "search_grid":
            fmtval = "[%s,%s]" % (val[0], "x".join("%g" % v for v in val[1]))
            return spec.argstr % fmtval

        if opt == "fixed_image_mask":
            if isdefined(self.inputs.moving_image_mask):
                return spec.argstr % ("[%s,%s]" % (val, self.inputs.moving_image_mask))

        return super(AI, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        return getattr(self, "_output")


class AverageAffineTransformInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, argstr="%d", mandatory=True, position=0, desc="image dimension (2 or 3)"
    )
    output_affine_transform = File(
        argstr="%s",
        mandatory=True,
        position=1,
        desc="Outputfname.txt: the name of the resulting transform.",
    )
    transforms = InputMultiObject(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=3,
        desc="transforms to average",
    )


class AverageAffineTransformOutputSpec(TraitedSpec):
    affine_transform = File(exists=True, desc="average transform file")


class AverageAffineTransform(ANTSCommand):
    """
    Examples
    --------
    >>> from nipype.interfaces.ants import AverageAffineTransform
    >>> avg = AverageAffineTransform()
    >>> avg.inputs.dimension = 3
    >>> avg.inputs.transforms = ['trans.mat', 'func_to_struct.mat']
    >>> avg.inputs.output_affine_transform = 'MYtemplatewarp.mat'
    >>> avg.cmdline
    'AverageAffineTransform 3 MYtemplatewarp.mat trans.mat func_to_struct.mat'

    """

    _cmd = "AverageAffineTransform"
    input_spec = AverageAffineTransformInputSpec
    output_spec = AverageAffineTransformOutputSpec

    def _format_arg(self, opt, spec, val):
        return super(AverageAffineTransform, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["affine_transform"] = os.path.abspath(
            self.inputs.output_affine_transform
        )
        return outputs


class AverageImagesInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, argstr="%d", mandatory=True, position=0, desc="image dimension (2 or 3)"
    )
    output_average_image = File(
        "average.nii",
        argstr="%s",
        position=1,
        usedefault=True,
        hash_files=False,
        desc="the name of the resulting image.",
    )
    normalize = traits.Bool(
        argstr="%d",
        mandatory=True,
        position=2,
        desc="Normalize: if true, the 2nd image is divided by its mean. "
        "This will select the largest image to average into.",
    )
    images = InputMultiObject(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=3,
        desc="image to apply transformation to (generally a coregistered functional)",
    )


class AverageImagesOutputSpec(TraitedSpec):
    output_average_image = File(exists=True, desc="average image file")


class AverageImages(ANTSCommand):
    """
    Examples
    --------
    >>> from nipype.interfaces.ants import AverageImages
    >>> avg = AverageImages()
    >>> avg.inputs.dimension = 3
    >>> avg.inputs.output_average_image = "average.nii.gz"
    >>> avg.inputs.normalize = True
    >>> avg.inputs.images = ['rc1s1.nii', 'rc1s1.nii']
    >>> avg.cmdline
    'AverageImages 3 average.nii.gz 1 rc1s1.nii rc1s1.nii'
    """

    _cmd = "AverageImages"
    input_spec = AverageImagesInputSpec
    output_spec = AverageImagesOutputSpec

    def _format_arg(self, opt, spec, val):
        return super(AverageImages, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["output_average_image"] = os.path.realpath(
            self.inputs.output_average_image
        )
        return outputs


class MultiplyImagesInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, argstr="%d", mandatory=True, position=0, desc="image dimension (2 or 3)"
    )
    first_input = File(
        argstr="%s", exists=True, mandatory=True, position=1, desc="image 1"
    )
    second_input = traits.Either(
        File(exists=True),
        traits.Float,
        argstr="%s",
        mandatory=True,
        position=2,
        desc="image 2 or multiplication weight",
    )
    output_product_image = File(
        argstr="%s",
        mandatory=True,
        position=3,
        desc="Outputfname.nii.gz: the name of the resulting image.",
    )


class MultiplyImagesOutputSpec(TraitedSpec):
    output_product_image = File(exists=True, desc="average image file")


class MultiplyImages(ANTSCommand):
    """
    Examples
    --------
    >>> from nipype.interfaces.ants import MultiplyImages
    >>> test = MultiplyImages()
    >>> test.inputs.dimension = 3
    >>> test.inputs.first_input = 'moving2.nii'
    >>> test.inputs.second_input = 0.25
    >>> test.inputs.output_product_image = "out.nii"
    >>> test.cmdline
    'MultiplyImages 3 moving2.nii 0.25 out.nii'
    """

    _cmd = "MultiplyImages"
    input_spec = MultiplyImagesInputSpec
    output_spec = MultiplyImagesOutputSpec

    def _format_arg(self, opt, spec, val):
        return super(MultiplyImages, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["output_product_image"] = os.path.abspath(
            self.inputs.output_product_image
        )
        return outputs


class CreateJacobianDeterminantImageInputSpec(ANTSCommandInputSpec):
    imageDimension = traits.Enum(
        3, 2, argstr="%d", mandatory=True, position=0, desc="image dimension (2 or 3)"
    )
    deformationField = File(
        argstr="%s",
        exists=True,
        mandatory=True,
        position=1,
        desc="deformation transformation file",
    )
    outputImage = File(argstr="%s", mandatory=True, position=2, desc="output filename")
    doLogJacobian = traits.Enum(
        0, 1, argstr="%d", position=3, desc="return the log jacobian"
    )
    useGeometric = traits.Enum(
        0, 1, argstr="%d", position=4, desc="return the geometric jacobian"
    )


class CreateJacobianDeterminantImageOutputSpec(TraitedSpec):
    jacobian_image = File(exists=True, desc="jacobian image")


class CreateJacobianDeterminantImage(ANTSCommand):
    """
    Examples
    --------
    >>> from nipype.interfaces.ants import CreateJacobianDeterminantImage
    >>> jacobian = CreateJacobianDeterminantImage()
    >>> jacobian.inputs.imageDimension = 3
    >>> jacobian.inputs.deformationField = 'ants_Warp.nii.gz'
    >>> jacobian.inputs.outputImage = 'out_name.nii.gz'
    >>> jacobian.cmdline
    'CreateJacobianDeterminantImage 3 ants_Warp.nii.gz out_name.nii.gz'
    """

    _cmd = "CreateJacobianDeterminantImage"
    input_spec = CreateJacobianDeterminantImageInputSpec
    output_spec = CreateJacobianDeterminantImageOutputSpec

    def _format_arg(self, opt, spec, val):
        return super(CreateJacobianDeterminantImage, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["jacobian_image"] = os.path.abspath(self.inputs.outputImage)
        return outputs


class AffineInitializerInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, usedefault=True, position=0, argstr="%s", desc="dimension"
    )
    fixed_image = File(
        exists=True, mandatory=True, position=1, argstr="%s", desc="reference image"
    )
    moving_image = File(
        exists=True, mandatory=True, position=2, argstr="%s", desc="moving image"
    )
    out_file = File(
        "transform.mat",
        usedefault=True,
        position=3,
        argstr="%s",
        desc="output transform file",
    )
    # Defaults in antsBrainExtraction.sh -> 15 0.1 0 10
    search_factor = traits.Float(
        15.0,
        usedefault=True,
        position=4,
        argstr="%f",
        desc="increments (degrees) for affine search",
    )
    radian_fraction = traits.Range(
        0.0,
        1.0,
        value=0.1,
        usedefault=True,
        position=5,
        argstr="%f",
        desc="search this arc +/- principal axes",
    )
    principal_axes = traits.Bool(
        False,
        usedefault=True,
        position=6,
        argstr="%d",
        desc="whether the rotation is searched around an initial principal axis alignment.",
    )
    local_search = traits.Int(
        10,
        usedefault=True,
        position=7,
        argstr="%d",
        desc=" determines if a local optimization is run at each search point for the set "
        "number of iterations",
    )


class AffineInitializerOutputSpec(TraitedSpec):
    out_file = File(desc="output transform file")


class AffineInitializer(ANTSCommand):
    """
    Initialize an affine transform (as in antsBrainExtraction.sh)

    >>> from nipype.interfaces.ants import AffineInitializer
    >>> init = AffineInitializer()
    >>> init.inputs.fixed_image = 'fixed1.nii'
    >>> init.inputs.moving_image = 'moving1.nii'
    >>> init.cmdline
    'antsAffineInitializer 3 fixed1.nii moving1.nii transform.mat 15.000000 0.100000 0 10'

    """

    _cmd = "antsAffineInitializer"
    input_spec = AffineInitializerInputSpec
    output_spec = AffineInitializerOutputSpec

    def _list_outputs(self):
        return {"out_file": os.path.abspath(self.inputs.out_file)}


class ComposeMultiTransformInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, argstr="%d", usedefault=True, position=0, desc="image dimension (2 or 3)"
    )
    output_transform = File(
        argstr="%s",
        position=1,
        name_source=["transforms"],
        name_template="%s_composed",
        keep_extension=True,
        desc="the name of the resulting transform.",
    )
    reference_image = File(
        argstr="%s",
        position=2,
        desc="Reference image (only necessary when output is warpfield)",
    )
    transforms = InputMultiObject(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=3,
        desc="transforms to average",
    )


class ComposeMultiTransformOutputSpec(TraitedSpec):
    output_transform = File(exists=True, desc="Composed transform file")


class ComposeMultiTransform(ANTSCommand):
    """
    Take a set of transformations and convert them to a single transformation matrix/warpfield.

    Examples
    --------
    >>> from nipype.interfaces.ants import ComposeMultiTransform
    >>> compose_transform = ComposeMultiTransform()
    >>> compose_transform.inputs.dimension = 3
    >>> compose_transform.inputs.transforms = ['struct_to_template.mat', 'func_to_struct.mat']
    >>> compose_transform.cmdline
    'ComposeMultiTransform 3 struct_to_template_composed.mat
    struct_to_template.mat func_to_struct.mat'

    """

    _cmd = "ComposeMultiTransform"
    input_spec = ComposeMultiTransformInputSpec
    output_spec = ComposeMultiTransformOutputSpec


class LabelGeometryInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        3, 2, argstr="%d", usedefault=True, position=0, desc="image dimension (2 or 3)"
    )
    label_image = File(
        argstr="%s",
        position=1,
        mandatory=True,
        desc="label image to use for extracting geometry measures",
    )
    intensity_image = File(
        value="[]",
        exists=True,
        argstr="%s",
        mandatory=True,
        usedefault=True,
        position=2,
        desc="Intensity image to extract values from. " "This is an optional input",
    )
    output_file = traits.Str(
        name_source=["label_image"],
        name_template="%s.csv",
        argstr="%s",
        position=3,
        desc="name of output file",
    )


class LabelGeometryOutputSpec(TraitedSpec):
    output_file = File(exists=True, desc="CSV file of geometry measures")


class LabelGeometry(ANTSCommand):
    """
    Extracts geometry measures using a label file and an optional image file

    Examples
    --------
    >>> from nipype.interfaces.ants import LabelGeometry
    >>> label_extract = LabelGeometry()
    >>> label_extract.inputs.dimension = 3
    >>> label_extract.inputs.label_image = 'atlas.nii.gz'
    >>> label_extract.cmdline
    'LabelGeometryMeasures 3 atlas.nii.gz [] atlas.csv'

    >>> label_extract.inputs.intensity_image = 'ants_Warp.nii.gz'
    >>> label_extract.cmdline
    'LabelGeometryMeasures 3 atlas.nii.gz ants_Warp.nii.gz atlas.csv'

    """

    _cmd = "LabelGeometryMeasures"
    input_spec = LabelGeometryInputSpec
    output_spec = LabelGeometryOutputSpec



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..segmentation import LaplacianThickness


def test_LaplacianThickness_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        dT=dict(
            argstr="%s",
            position=6,
            requires=["prior_thickness"],
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        input_gm=dict(
            argstr="%s",
            copyfile=True,
            extensions=None,
            mandatory=True,
            position=2,
        ),
        input_wm=dict(
            argstr="%s",
            copyfile=True,
            extensions=None,
            mandatory=True,
            position=1,
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        output_image=dict(
            argstr="%s",
            hash_files=False,
            keep_extension=True,
            name_source=["input_wm"],
            name_template="%s_thickness",
            position=3,
        ),
        prior_thickness=dict(
            argstr="%s",
            position=5,
            requires=["smooth_param"],
        ),
        smooth_param=dict(
            argstr="%s",
            position=4,
        ),
        sulcus_prior=dict(
            argstr="%s",
            position=7,
            requires=["dT"],
        ),
        tolerance=dict(
            argstr="%s",
            position=8,
            requires=["sulcus_prior"],
        ),
    )
    inputs = LaplacianThickness.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_LaplacianThickness_outputs():
    output_map = dict(
        output_image=dict(
            extensions=None,
        ),
    )
    outputs = LaplacianThickness.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..segmentation import BRAINSTalairach


def test_BRAINSTalairach_inputs():
    input_map = dict(
        AC=dict(
            argstr="--AC %s",
            sep=",",
        ),
        ACisIndex=dict(
            argstr="--ACisIndex ",
        ),
        IRP=dict(
            argstr="--IRP %s",
            sep=",",
        ),
        IRPisIndex=dict(
            argstr="--IRPisIndex ",
        ),
        PC=dict(
            argstr="--PC %s",
            sep=",",
        ),
        PCisIndex=dict(
            argstr="--PCisIndex ",
        ),
        SLA=dict(
            argstr="--SLA %s",
            sep=",",
        ),
        SLAisIndex=dict(
            argstr="--SLAisIndex ",
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputVolume=dict(
            argstr="--inputVolume %s",
            extensions=None,
        ),
        outputBox=dict(
            argstr="--outputBox %s",
            hash_files=False,
        ),
        outputGrid=dict(
            argstr="--outputGrid %s",
            hash_files=False,
        ),
    )
    inputs = BRAINSTalairach.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_BRAINSTalairach_outputs():
    output_map = dict(
        outputBox=dict(
            extensions=None,
        ),
        outputGrid=dict(
            extensions=None,
        ),
    )
    outputs = BRAINSTalairach.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..brainsuite import Cortex


def test_Cortex_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        computeGCBoundary=dict(
            argstr="-g",
        ),
        computeWGBoundary=dict(
            argstr="-w",
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        includeAllSubcorticalAreas=dict(
            argstr="-a",
            usedefault=True,
        ),
        inputHemisphereLabelFile=dict(
            argstr="-h %s",
            extensions=None,
            mandatory=True,
        ),
        inputTissueFractionFile=dict(
            argstr="-f %s",
            extensions=None,
            mandatory=True,
        ),
        outputCerebrumMask=dict(
            argstr="-o %s",
            extensions=None,
            genfile=True,
        ),
        timer=dict(
            argstr="--timer",
        ),
        tissueFractionThreshold=dict(
            argstr="-p %f",
            usedefault=True,
        ),
        verbosity=dict(
            argstr="-v %d",
        ),
    )
    inputs = Cortex.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Cortex_outputs():
    output_map = dict(
        outputCerebrumMask=dict(
            extensions=None,
        ),
    )
    outputs = Cortex.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import os.path as op

from ...utils.filemanip import split_filename
from ..base import (
    CommandLineInputSpec,
    CommandLine,
    traits,
    TraitedSpec,
    File,
    InputMultiPath,
    isdefined,
)


class MRConvertInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="voxel-order data filename",
    )
    out_filename = File(genfile=True, argstr="%s", position=-1, desc="Output filename")
    extract_at_axis = traits.Enum(
        1,
        2,
        3,
        argstr="-coord %s",
        position=1,
        desc='"Extract data only at the coordinates specified. This option specifies the Axis. Must be used in conjunction with extract_at_coordinate.',
    )
    extract_at_coordinate = traits.List(
        traits.Float,
        argstr="%s",
        sep=",",
        position=2,
        minlen=1,
        maxlen=3,
        desc='"Extract data only at the coordinates specified. This option specifies the coordinates. Must be used in conjunction with extract_at_axis. Three comma-separated numbers giving the size of each voxel in mm.',
    )
    voxel_dims = traits.List(
        traits.Float,
        argstr="-vox %s",
        sep=",",
        position=3,
        minlen=3,
        maxlen=3,
        desc="Three comma-separated numbers giving the size of each voxel in mm.",
    )
    output_datatype = traits.Enum(
        "nii",
        "float",
        "char",
        "short",
        "int",
        "long",
        "double",
        argstr="-output %s",
        position=2,
        desc='"i.e. Bfloat". Can be "char", "short", "int", "long", "float" or "double"',
    )  # , usedefault=True)
    extension = traits.Enum(
        "mif",
        "nii",
        "float",
        "char",
        "short",
        "int",
        "long",
        "double",
        position=2,
        desc='"i.e. Bfloat". Can be "char", "short", "int", "long", "float" or "double"',
        usedefault=True,
    )
    layout = traits.Enum(
        "nii",
        "float",
        "char",
        "short",
        "int",
        "long",
        "double",
        argstr="-output %s",
        position=2,
        desc="specify the layout of the data in memory. The actual layout produced will depend on whether the output image format can support it.",
    )
    resample = traits.Float(
        argstr="-scale %d",
        position=3,
        units="mm",
        desc="Apply scaling to the intensity values.",
    )
    offset_bias = traits.Float(
        argstr="-scale %d",
        position=3,
        units="mm",
        desc="Apply offset to the intensity values.",
    )
    replace_NaN_with_zero = traits.Bool(
        argstr="-zero", position=3, desc="Replace all NaN values with zero."
    )
    prs = traits.Bool(
        argstr="-prs",
        position=3,
        desc="Assume that the DW gradients are specified in the PRS frame (Siemens DICOM only).",
    )


class MRConvertOutputSpec(TraitedSpec):
    converted = File(exists=True, desc="path/name of 4D volume in voxel order")


class MRConvert(CommandLine):
    """
    Perform conversion between different file types and optionally extract a subset of the input image.

    If used correctly, this program can be a very useful workhorse.
    In addition to converting images between different formats, it can
    be used to extract specific studies from a data set, extract a specific
    region of interest, flip the images, or to scale the intensity of the images.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> mrconvert = mrt.MRConvert()
    >>> mrconvert.inputs.in_file = 'dwi_FA.mif'
    >>> mrconvert.inputs.out_filename = 'dwi_FA.nii'
    >>> mrconvert.run()                                 # doctest: +SKIP
    """

    _cmd = "mrconvert"
    input_spec = MRConvertInputSpec
    output_spec = MRConvertOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["converted"] = self.inputs.out_filename
        if not isdefined(outputs["converted"]):
            outputs["converted"] = op.abspath(self._gen_outfilename())
        else:
            outputs["converted"] = op.abspath(outputs["converted"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        if isdefined(self.inputs.out_filename):
            outname = self.inputs.out_filename
        else:
            outname = name + "_mrconvert." + self.inputs.extension
        return outname


class DWI2TensorInputSpec(CommandLineInputSpec):
    in_file = InputMultiPath(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Diffusion-weighted images",
    )
    out_filename = File(
        name_template="%s_tensor.mif",
        name_source="in_file",
        output_name="tensor",
        argstr="%s",
        desc="Output tensor filename",
        position=-1,
    )
    encoding_file = File(
        argstr="-grad %s",
        position=2,
        desc=(
            "Encoding file supplied as a 4xN text file with "
            "each line is in the format [ X Y Z b ], where "
            "[ X Y Z ] describe the direction of the applied "
            "gradient, and b gives the b-value in units "
            "(1000 s/mm^2). See FSL2MRTrix()"
        ),
    )
    ignore_slice_by_volume = traits.List(
        traits.Int,
        argstr="-ignoreslices %s",
        sep=" ",
        position=2,
        minlen=2,
        maxlen=2,
        desc=(
            "Requires two values (i.e. [34 "
            "1] for [Slice Volume] Ignores "
            "the image slices specified "
            "when computing the tensor. "
            "Slice here means the z "
            "coordinate of the slice to be "
            "ignored."
        ),
    )
    ignore_volumes = traits.List(
        traits.Int,
        argstr="-ignorevolumes %s",
        sep=" ",
        position=2,
        minlen=1,
        desc=(
            "Requires two values (i.e. [2 5 6] for "
            "[Volumes] Ignores the image volumes "
            "specified when computing the tensor."
        ),
    )
    mask = File(
        exists=True,
        argstr="-mask %s",
        desc="Only perform computation within the specified binary brain mask image.",
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc=("Do not display information messages or progress " "status."),
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")


class DWI2TensorOutputSpec(TraitedSpec):
    tensor = File(exists=True, desc="path/name of output diffusion tensor image")


class DWI2Tensor(CommandLine):
    """
    Converts diffusion-weighted images to tensor images.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> dwi2tensor = mrt.DWI2Tensor()
    >>> dwi2tensor.inputs.in_file = 'dwi.mif'
    >>> dwi2tensor.inputs.encoding_file = 'encoding.txt'
    >>> dwi2tensor.cmdline
    'dwi2tensor -grad encoding.txt dwi.mif dwi_tensor.mif'
    >>> dwi2tensor.run()                                   # doctest: +SKIP
    """

    _cmd = "dwi2tensor"
    input_spec = DWI2TensorInputSpec
    output_spec = DWI2TensorOutputSpec


class Tensor2VectorInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Diffusion tensor image",
    )
    out_filename = File(
        genfile=True, argstr="%s", position=-1, desc="Output vector filename"
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc="Do not display information messages or progress status.",
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")


class Tensor2VectorOutputSpec(TraitedSpec):
    vector = File(
        exists=True,
        desc="the output image of the major eigenvectors of the diffusion tensor image.",
    )


class Tensor2Vector(CommandLine):
    """
    Generates a map of the major eigenvectors of the tensors in each voxel.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> tensor2vector = mrt.Tensor2Vector()
    >>> tensor2vector.inputs.in_file = 'dwi_tensor.mif'
    >>> tensor2vector.run()                             # doctest: +SKIP
    """

    _cmd = "tensor2vector"
    input_spec = Tensor2VectorInputSpec
    output_spec = Tensor2VectorOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["vector"] = self.inputs.out_filename
        if not isdefined(outputs["vector"]):
            outputs["vector"] = op.abspath(self._gen_outfilename())
        else:
            outputs["vector"] = op.abspath(outputs["vector"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_vector.mif"


class Tensor2FractionalAnisotropyInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Diffusion tensor image",
    )
    out_filename = File(
        genfile=True,
        argstr="%s",
        position=-1,
        desc="Output Fractional Anisotropy filename",
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc="Do not display information messages or progress status.",
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")


class Tensor2FractionalAnisotropyOutputSpec(TraitedSpec):
    FA = File(
        exists=True,
        desc="the output image of the major eigenvectors of the diffusion tensor image.",
    )


class Tensor2FractionalAnisotropy(CommandLine):
    """
    Generates a map of the fractional anisotropy in each voxel.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> tensor2FA = mrt.Tensor2FractionalAnisotropy()
    >>> tensor2FA.inputs.in_file = 'dwi_tensor.mif'
    >>> tensor2FA.run()                                 # doctest: +SKIP
    """

    _cmd = "tensor2FA"
    input_spec = Tensor2FractionalAnisotropyInputSpec
    output_spec = Tensor2FractionalAnisotropyOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["FA"] = self.inputs.out_filename
        if not isdefined(outputs["FA"]):
            outputs["FA"] = op.abspath(self._gen_outfilename())
        else:
            outputs["FA"] = op.abspath(outputs["FA"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_FA.mif"


class Tensor2ApparentDiffusionInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Diffusion tensor image",
    )
    out_filename = File(
        genfile=True,
        argstr="%s",
        position=-1,
        desc="Output Fractional Anisotropy filename",
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc="Do not display information messages or progress status.",
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")


class Tensor2ApparentDiffusionOutputSpec(TraitedSpec):
    ADC = File(
        exists=True,
        desc="the output image of the major eigenvectors of the diffusion tensor image.",
    )


class Tensor2ApparentDiffusion(CommandLine):
    """
    Generates a map of the apparent diffusion coefficient (ADC) in each voxel

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> tensor2ADC = mrt.Tensor2ApparentDiffusion()
    >>> tensor2ADC.inputs.in_file = 'dwi_tensor.mif'
    >>> tensor2ADC.run()                                # doctest: +SKIP
    """

    _cmd = "tensor2ADC"
    input_spec = Tensor2ApparentDiffusionInputSpec
    output_spec = Tensor2ApparentDiffusionOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["ADC"] = self.inputs.out_filename
        if not isdefined(outputs["ADC"]):
            outputs["ADC"] = op.abspath(self._gen_outfilename())
        else:
            outputs["ADC"] = op.abspath(outputs["ADC"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_ADC.mif"


class MRMultiplyInputSpec(CommandLineInputSpec):
    in_files = InputMultiPath(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Input images to be multiplied",
    )
    out_filename = File(
        genfile=True, argstr="%s", position=-1, desc="Output image filename"
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc="Do not display information messages or progress status.",
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")


class MRMultiplyOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output image of the multiplication")


class MRMultiply(CommandLine):
    """
    Multiplies two images.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> MRmult = mrt.MRMultiply()
    >>> MRmult.inputs.in_files = ['dwi.mif', 'dwi_WMProb.mif']
    >>> MRmult.run()                                             # doctest: +SKIP
    """

    _cmd = "mrmult"
    input_spec = MRMultiplyInputSpec
    output_spec = MRMultiplyOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self.inputs.out_filename
        if not isdefined(outputs["out_file"]):
            outputs["out_file"] = op.abspath(self._gen_outfilename())
        else:
            outputs["out_file"] = op.abspath(outputs["out_file"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_files[0])
        return name + "_MRMult.mif"


class MRTrixViewerInputSpec(CommandLineInputSpec):
    in_files = InputMultiPath(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Input images to be viewed",
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc="Do not display information messages or progress status.",
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")


class MRTrixViewerOutputSpec(TraitedSpec):
    pass


class MRTrixViewer(CommandLine):
    """
    Loads the input images in the MRTrix Viewer.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> MRview = mrt.MRTrixViewer()
    >>> MRview.inputs.in_files = 'dwi.mif'
    >>> MRview.run()                                    # doctest: +SKIP
    """

    _cmd = "mrview"
    input_spec = MRTrixViewerInputSpec
    output_spec = MRTrixViewerOutputSpec

    def _list_outputs(self):
        return


class MRTrixInfoInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Input images to be read",
    )


class MRTrixInfoOutputSpec(TraitedSpec):
    pass


class MRTrixInfo(CommandLine):
    """
    Prints out relevant header information found in the image specified.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> MRinfo = mrt.MRTrixInfo()
    >>> MRinfo.inputs.in_file = 'dwi.mif'
    >>> MRinfo.run()                                    # doctest: +SKIP
    """

    _cmd = "mrinfo"
    input_spec = MRTrixInfoInputSpec
    output_spec = MRTrixInfoOutputSpec

    def _list_outputs(self):
        return


class GenerateWhiteMatterMaskInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-3,
        desc="Diffusion-weighted images",
    )
    binary_mask = File(
        exists=True, argstr="%s", mandatory=True, position=-2, desc="Binary brain mask"
    )
    out_WMProb_filename = File(
        genfile=True,
        argstr="%s",
        position=-1,
        desc="Output WM probability image filename",
    )
    encoding_file = File(
        exists=True,
        argstr="-grad %s",
        mandatory=True,
        position=1,
        desc="Gradient encoding, supplied as a 4xN text file with each line is in the format [ X Y Z b ], where [ X Y Z ] describe the direction of the applied gradient, and b gives the b-value in units (1000 s/mm^2). See FSL2MRTrix",
    )
    noise_level_margin = traits.Float(
        argstr="-margin %s",
        desc="Specify the width of the margin on either side of the image to be used to estimate the noise level (default = 10)",
    )


class GenerateWhiteMatterMaskOutputSpec(TraitedSpec):
    WMprobabilitymap = File(exists=True, desc="WMprobabilitymap")


class GenerateWhiteMatterMask(CommandLine):
    """
    Generates a white matter probability mask from the DW images.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> genWM = mrt.GenerateWhiteMatterMask()
    >>> genWM.inputs.in_file = 'dwi.mif'
    >>> genWM.inputs.encoding_file = 'encoding.txt'
    >>> genWM.run()                                     # doctest: +SKIP
    """

    _cmd = "gen_WM_mask"
    input_spec = GenerateWhiteMatterMaskInputSpec
    output_spec = GenerateWhiteMatterMaskOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["WMprobabilitymap"] = op.abspath(self._gen_outfilename())
        return outputs

    def _gen_filename(self, name):
        if name == "out_WMProb_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_WMProb.mif"


class ErodeInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Input mask image to be eroded",
    )
    out_filename = File(
        genfile=True, argstr="%s", position=-1, desc="Output image filename"
    )
    number_of_passes = traits.Int(
        argstr="-npass %s", desc="the number of passes (default: 1)"
    )
    dilate = traits.Bool(
        argstr="-dilate", position=1, desc="Perform dilation rather than erosion"
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc="Do not display information messages or progress status.",
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")


class ErodeOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output image")


class Erode(CommandLine):
    """
    Erode (or dilates) a mask (i.e. binary) image

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> erode = mrt.Erode()
    >>> erode.inputs.in_file = 'mask.mif'
    >>> erode.run()                                     # doctest: +SKIP
    """

    _cmd = "erode"
    input_spec = ErodeInputSpec
    output_spec = ErodeOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self.inputs.out_filename
        if not isdefined(outputs["out_file"]):
            outputs["out_file"] = op.abspath(self._gen_outfilename())
        else:
            outputs["out_file"] = op.abspath(outputs["out_file"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_erode.mif"


class ThresholdInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="The input image to be thresholded",
    )
    out_filename = File(
        genfile=True, argstr="%s", position=-1, desc="The output binary image mask."
    )
    absolute_threshold_value = traits.Float(
        argstr="-abs %s", desc="Specify threshold value as absolute intensity."
    )
    percentage_threshold_value = traits.Float(
        argstr="-percent %s",
        desc="Specify threshold value as a percentage of the peak intensity in the input image.",
    )
    invert = traits.Bool(argstr="-invert", position=1, desc="Invert output binary mask")
    replace_zeros_with_NaN = traits.Bool(
        argstr="-nan", position=1, desc="Replace all zero values with NaN"
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc="Do not display information messages or progress status.",
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")


class ThresholdOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="The output binary image mask.")


class Threshold(CommandLine):
    """
    Create bitwise image by thresholding image intensity.

    By default, the threshold level is determined using a histogram analysis
    to cut out the background. Otherwise, the threshold intensity can be
    specified using command line options.
    Note that only the first study is used for thresholding.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> thresh = mrt.Threshold()
    >>> thresh.inputs.in_file = 'wm_mask.mif'
    >>> thresh.run()                                             # doctest: +SKIP
    """

    _cmd = "threshold"
    input_spec = ThresholdInputSpec
    output_spec = ThresholdOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self.inputs.out_filename
        if not isdefined(outputs["out_file"]):
            outputs["out_file"] = op.abspath(self._gen_outfilename())
        else:
            outputs["out_file"] = op.abspath(outputs["out_file"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_thresh.mif"


class MedianFilter3DInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Input images to be smoothed",
    )
    out_filename = File(
        genfile=True, argstr="%s", position=-1, desc="Output image filename"
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc="Do not display information messages or progress status.",
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")


class MedianFilter3DOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output image")


class MedianFilter3D(CommandLine):
    """
    Smooth images using a 3x3x3 median filter.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> median3d = mrt.MedianFilter3D()
    >>> median3d.inputs.in_file = 'mask.mif'
    >>> median3d.run()                                  # doctest: +SKIP
    """

    _cmd = "median3D"
    input_spec = MedianFilter3DInputSpec
    output_spec = MedianFilter3DOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self.inputs.out_filename
        if not isdefined(outputs["out_file"]):
            outputs["out_file"] = op.abspath(self._gen_outfilename())
        else:
            outputs["out_file"] = op.abspath(outputs["out_file"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_median3D.mif"


class MRTransformInputSpec(CommandLineInputSpec):
    in_files = InputMultiPath(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Input images to be transformed",
    )
    out_filename = File(genfile=True, argstr="%s", position=-1, desc="Output image")
    invert = traits.Bool(
        argstr="-inverse",
        position=1,
        desc="Invert the specified transform before using it",
    )
    linear_transform = File(
        exists=True,
        argstr="-linear %s",
        position=1,
        desc=(
            "Specify a linear transform to apply, in the form of a 3x4 or 4x4 ascii file. "
            "Note the standard reverse convention is used, "
            "where the transform maps points in the template image to the moving image. "
            "Note that the reverse convention is still assumed even if no -template image is supplied."
        ),
    )
    replace_transform = traits.Bool(
        argstr="-replace",
        position=1,
        desc="replace the current transform by that specified, rather than applying it to the current transform",
    )
    transformation_file = File(
        exists=True,
        argstr="-transform %s",
        position=1,
        desc="The transform to apply, in the form of a 4x4 ascii file.",
    )
    template_image = File(
        exists=True,
        argstr="-template %s",
        position=1,
        desc="Reslice the input image to match the specified template image.",
    )
    reference_image = File(
        exists=True,
        argstr="-reference %s",
        position=1,
        desc="in case the transform supplied maps from the input image onto a reference image, use this option to specify the reference. Note that this implicitly sets the -replace option.",
    )
    flip_x = traits.Bool(
        argstr="-flipx",
        position=1,
        desc="assume the transform is supplied assuming a coordinate system with the x-axis reversed relative to the MRtrix convention (i.e. x increases from right to left). This is required to handle transform matrices produced by FSL's FLIRT command. This is only used in conjunction with the -reference option.",
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc="Do not display information messages or progress status.",
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")


class MRTransformOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output image of the transformation")


class MRTransform(CommandLine):
    """
    Apply spatial transformations or reslice images

    Example
    -------

    >>> MRxform = MRTransform()
    >>> MRxform.inputs.in_files = 'anat_coreg.mif'
    >>> MRxform.run()                                   # doctest: +SKIP
    """

    _cmd = "mrtransform"
    input_spec = MRTransformInputSpec
    output_spec = MRTransformOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self.inputs.out_filename
        if not isdefined(outputs["out_file"]):
            outputs["out_file"] = op.abspath(self._gen_outfilename())
        else:
            outputs["out_file"] = op.abspath(outputs["out_file"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_files[0])
        return name + "_MRTransform.mif"



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import ApplyWarp


def test_ApplyWarp_inputs():
    input_map = dict(
        abswarp=dict(
            argstr="--abs",
            xor=["relwarp"],
        ),
        args=dict(
            argstr="%s",
        ),
        datatype=dict(
            argstr="--datatype=%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        field_file=dict(
            argstr="--warp=%s",
            extensions=None,
        ),
        in_file=dict(
            argstr="--in=%s",
            extensions=None,
            mandatory=True,
            position=0,
        ),
        interp=dict(
            argstr="--interp=%s",
            position=-2,
        ),
        mask_file=dict(
            argstr="--mask=%s",
            extensions=None,
        ),
        out_file=dict(
            argstr="--out=%s",
            extensions=None,
            genfile=True,
            hash_files=False,
            position=2,
        ),
        output_type=dict(),
        postmat=dict(
            argstr="--postmat=%s",
            extensions=None,
        ),
        premat=dict(
            argstr="--premat=%s",
            extensions=None,
        ),
        ref_file=dict(
            argstr="--ref=%s",
            extensions=None,
            mandatory=True,
            position=1,
        ),
        relwarp=dict(
            argstr="--rel",
            position=-1,
            xor=["abswarp"],
        ),
        superlevel=dict(
            argstr="--superlevel=%s",
        ),
        supersample=dict(
            argstr="--super",
        ),
    )
    inputs = ApplyWarp.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ApplyWarp_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = ApplyWarp.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..confounds import ComputeDVARS


def test_ComputeDVARS_inputs():
    input_map = dict(
        figdpi=dict(
            usedefault=True,
        ),
        figformat=dict(
            usedefault=True,
        ),
        figsize=dict(
            usedefault=True,
        ),
        in_file=dict(
            extensions=None,
            mandatory=True,
        ),
        in_mask=dict(
            extensions=None,
            mandatory=True,
        ),
        intensity_normalization=dict(
            usedefault=True,
        ),
        remove_zerovariance=dict(
            usedefault=True,
        ),
        save_all=dict(
            usedefault=True,
        ),
        save_nstd=dict(
            usedefault=True,
        ),
        save_plot=dict(
            usedefault=True,
        ),
        save_std=dict(
            usedefault=True,
        ),
        save_vxstd=dict(
            usedefault=True,
        ),
        series_tr=dict(),
    )
    inputs = ComputeDVARS.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ComputeDVARS_outputs():
    output_map = dict(
        avg_nstd=dict(),
        avg_std=dict(),
        avg_vxstd=dict(),
        fig_nstd=dict(
            extensions=None,
        ),
        fig_std=dict(
            extensions=None,
        ),
        fig_vxstd=dict(
            extensions=None,
        ),
        out_all=dict(
            extensions=None,
        ),
        out_nstd=dict(
            extensions=None,
        ),
        out_std=dict(
            extensions=None,
        ),
        out_vxstd=dict(
            extensions=None,
        ),
    )
    outputs = ComputeDVARS.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..cmtk import CreateMatrix


def test_CreateMatrix_inputs():
    input_map = dict(
        count_region_intersections=dict(
            usedefault=True,
        ),
        out_endpoint_array_name=dict(
            extensions=None,
            genfile=True,
        ),
        out_fiber_length_std_matrix_mat_file=dict(
            extensions=None,
            genfile=True,
        ),
        out_intersection_matrix_mat_file=dict(
            extensions=None,
            genfile=True,
        ),
        out_matrix_file=dict(
            extensions=None,
            genfile=True,
        ),
        out_matrix_mat_file=dict(
            extensions=None,
            usedefault=True,
        ),
        out_mean_fiber_length_matrix_mat_file=dict(
            extensions=None,
            genfile=True,
        ),
        out_median_fiber_length_matrix_mat_file=dict(
            extensions=None,
            genfile=True,
        ),
        resolution_network_file=dict(
            extensions=None,
            mandatory=True,
        ),
        roi_file=dict(
            extensions=None,
            mandatory=True,
        ),
        tract_file=dict(
            extensions=None,
            mandatory=True,
        ),
    )
    inputs = CreateMatrix.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_CreateMatrix_outputs():
    output_map = dict(
        endpoint_file=dict(
            extensions=None,
        ),
        endpoint_file_mm=dict(
            extensions=None,
        ),
        fiber_label_file=dict(
            extensions=None,
        ),
        fiber_labels_noorphans=dict(
            extensions=None,
        ),
        fiber_length_file=dict(
            extensions=None,
        ),
        fiber_length_std_matrix_mat_file=dict(
            extensions=None,
        ),
        filtered_tractographies=dict(),
        filtered_tractography=dict(
            extensions=None,
        ),
        filtered_tractography_by_intersections=dict(
            extensions=None,
        ),
        intersection_matrix_file=dict(
            extensions=None,
        ),
        intersection_matrix_mat_file=dict(
            extensions=None,
        ),
        matlab_matrix_files=dict(),
        matrix_file=dict(
            extensions=None,
        ),
        matrix_files=dict(),
        matrix_mat_file=dict(
            extensions=None,
        ),
        mean_fiber_length_matrix_mat_file=dict(
            extensions=None,
        ),
        median_fiber_length_matrix_mat_file=dict(
            extensions=None,
        ),
        stats_file=dict(
            extensions=None,
        ),
    )
    outputs = CreateMatrix.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..petstandarduptakevaluecomputation import PETStandardUptakeValueComputation


def test_PETStandardUptakeValueComputation_inputs():
    input_map = dict(
        OutputLabel=dict(
            argstr="--OutputLabel %s",
        ),
        OutputLabelValue=dict(
            argstr="--OutputLabelValue %s",
        ),
        SUVMax=dict(
            argstr="--SUVMax %s",
        ),
        SUVMean=dict(
            argstr="--SUVMean %s",
        ),
        SUVMin=dict(
            argstr="--SUVMin %s",
        ),
        args=dict(
            argstr="%s",
        ),
        color=dict(
            argstr="--color %s",
            extensions=None,
        ),
        csvFile=dict(
            argstr="--csvFile %s",
            hash_files=False,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        labelMap=dict(
            argstr="--labelMap %s",
            extensions=None,
        ),
        petDICOMPath=dict(
            argstr="--petDICOMPath %s",
        ),
        petVolume=dict(
            argstr="--petVolume %s",
            extensions=None,
        ),
    )
    inputs = PETStandardUptakeValueComputation.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_PETStandardUptakeValueComputation_outputs():
    output_map = dict(
        csvFile=dict(
            extensions=None,
        ),
    )
    outputs = PETStandardUptakeValueComputation.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""
Import lib and class meta programming utilities.
"""
import inspect
import importlib

from ..interfaces.base import Interface


def import_module(module_path):
    """Import any module to the global Python environment.
       The module_path argument specifies what module to import in
       absolute or relative terms (e.g. either pkg.mod or ..mod).
       If the name is specified in relative terms, then the package argument
       must be set to the name of the package which is to act as the anchor
       for resolving the package name (e.g. import_module('..mod',
       'pkg.subpkg') will import pkg.mod).

    Parameters
    ----------
    module_path: str
        Path to the module to be imported

    Returns
    -------
    The specified module will be inserted into sys.modules and returned.
    """
    try:
        mod = importlib.import_module(module_path)
    except:
        raise ImportError("Error when importing object {}.".format(module_path))
    else:
        return mod


def list_interfaces(module):
    """Return a list with the names of the Interface subclasses inside
    the given module.
    """
    iface_names = []
    for k, v in sorted(list(module.__dict__.items())):
        if inspect.isclass(v) and issubclass(v, Interface):
            iface_names.append(k)
    return iface_names



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..registration import Registration


def test_Registration_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        collapse_output_transforms=dict(
            argstr="--collapse-output-transforms %d",
            usedefault=True,
        ),
        convergence_threshold=dict(
            requires=["number_of_iterations"],
            usedefault=True,
        ),
        convergence_window_size=dict(
            requires=["convergence_threshold"],
            usedefault=True,
        ),
        dimension=dict(
            argstr="--dimensionality %d",
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fixed_image=dict(
            mandatory=True,
        ),
        fixed_image_mask=dict(
            argstr="%s",
            extensions=None,
            max_ver="2.1.0",
            xor=["fixed_image_masks"],
        ),
        fixed_image_masks=dict(
            min_ver="2.2.0",
            xor=["fixed_image_mask"],
        ),
        float=dict(
            argstr="--float %d",
        ),
        initial_moving_transform=dict(
            argstr="%s",
            xor=["initial_moving_transform_com"],
        ),
        initial_moving_transform_com=dict(
            argstr="%s",
            xor=["initial_moving_transform"],
        ),
        initialize_transforms_per_stage=dict(
            argstr="--initialize-transforms-per-stage %d",
            usedefault=True,
        ),
        interpolation=dict(
            argstr="%s",
            usedefault=True,
        ),
        interpolation_parameters=dict(),
        invert_initial_moving_transform=dict(
            requires=["initial_moving_transform"],
            xor=["initial_moving_transform_com"],
        ),
        metric=dict(
            mandatory=True,
        ),
        metric_item_trait=dict(),
        metric_stage_trait=dict(),
        metric_weight=dict(
            mandatory=True,
            requires=["metric"],
            usedefault=True,
        ),
        metric_weight_item_trait=dict(
            usedefault=True,
        ),
        metric_weight_stage_trait=dict(),
        moving_image=dict(
            mandatory=True,
        ),
        moving_image_mask=dict(
            extensions=None,
            max_ver="2.1.0",
            requires=["fixed_image_mask"],
            xor=["moving_image_masks"],
        ),
        moving_image_masks=dict(
            min_ver="2.2.0",
            xor=["moving_image_mask"],
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        number_of_iterations=dict(),
        output_inverse_warped_image=dict(
            hash_files=False,
            requires=["output_warped_image"],
        ),
        output_transform_prefix=dict(
            argstr="%s",
            usedefault=True,
        ),
        output_warped_image=dict(
            hash_files=False,
        ),
        radius_bins_item_trait=dict(
            usedefault=True,
        ),
        radius_bins_stage_trait=dict(),
        radius_or_number_of_bins=dict(
            requires=["metric_weight"],
            usedefault=True,
        ),
        random_seed=dict(
            argstr="--random-seed %d",
            min_ver="2.3.0",
        ),
        restore_state=dict(
            argstr="--restore-state %s",
            extensions=None,
        ),
        restrict_deformation=dict(),
        sampling_percentage=dict(
            requires=["sampling_strategy"],
        ),
        sampling_percentage_item_trait=dict(),
        sampling_percentage_stage_trait=dict(),
        sampling_strategy=dict(
            requires=["metric_weight"],
        ),
        sampling_strategy_item_trait=dict(),
        sampling_strategy_stage_trait=dict(),
        save_state=dict(
            argstr="--save-state %s",
            extensions=None,
        ),
        shrink_factors=dict(
            mandatory=True,
        ),
        sigma_units=dict(
            requires=["smoothing_sigmas"],
        ),
        smoothing_sigmas=dict(
            mandatory=True,
        ),
        transform_parameters=dict(),
        transforms=dict(
            argstr="%s",
            mandatory=True,
        ),
        use_estimate_learning_rate_once=dict(),
        use_histogram_matching=dict(
            usedefault=True,
        ),
        verbose=dict(
            argstr="-v",
            nohash=True,
            usedefault=True,
        ),
        winsorize_lower_quantile=dict(
            argstr="%s",
            usedefault=True,
        ),
        winsorize_upper_quantile=dict(
            argstr="%s",
            usedefault=True,
        ),
        write_composite_transform=dict(
            argstr="--write-composite-transform %d",
            usedefault=True,
        ),
    )
    inputs = Registration.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Registration_outputs():
    output_map = dict(
        composite_transform=dict(
            extensions=None,
        ),
        elapsed_time=dict(),
        forward_invert_flags=dict(),
        forward_transforms=dict(),
        inverse_composite_transform=dict(
            extensions=None,
        ),
        inverse_warped_image=dict(
            extensions=None,
        ),
        metric_value=dict(),
        reverse_forward_invert_flags=dict(),
        reverse_forward_transforms=dict(),
        reverse_invert_flags=dict(),
        reverse_transforms=dict(),
        save_state=dict(
            extensions=None,
        ),
        warped_image=dict(
            extensions=None,
        ),
    )
    outputs = Registration.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import NetCorr


def test_NetCorr_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fish_z=dict(
            argstr="-fish_z",
        ),
        ignore_LT=dict(
            argstr="-ignore_LT",
        ),
        in_file=dict(
            argstr="-inset %s",
            extensions=None,
            mandatory=True,
        ),
        in_rois=dict(
            argstr="-in_rois %s",
            extensions=None,
            mandatory=True,
        ),
        mask=dict(
            argstr="-mask %s",
            extensions=None,
        ),
        nifti=dict(
            argstr="-nifti",
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_file=dict(
            argstr="-prefix %s",
            extensions=None,
            name_source="in_file",
            name_template="%s_netcorr",
            position=1,
        ),
        output_mask_nonnull=dict(
            argstr="-output_mask_nonnull",
        ),
        outputtype=dict(),
        part_corr=dict(
            argstr="-part_corr",
        ),
        push_thru_many_zeros=dict(
            argstr="-push_thru_many_zeros",
        ),
        ts_indiv=dict(
            argstr="-ts_indiv",
        ),
        ts_label=dict(
            argstr="-ts_label",
        ),
        ts_out=dict(
            argstr="-ts_out",
        ),
        ts_wb_Z=dict(
            argstr="-ts_wb_Z",
        ),
        ts_wb_corr=dict(
            argstr="-ts_wb_corr",
        ),
        ts_wb_strlabel=dict(
            argstr="-ts_wb_strlabel",
        ),
        weight_ts=dict(
            argstr="-weight_ts %s",
            extensions=None,
        ),
    )
    inputs = NetCorr.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_NetCorr_outputs():
    output_map = dict(
        out_corr_maps=dict(),
        out_corr_matrix=dict(
            extensions=None,
        ),
    )
    outputs = NetCorr.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import MCFLIRT


def test_MCFLIRT_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        bins=dict(
            argstr="-bins %d",
        ),
        cost=dict(
            argstr="-cost %s",
        ),
        dof=dict(
            argstr="-dof %d",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-in %s",
            extensions=None,
            mandatory=True,
            position=0,
        ),
        init=dict(
            argstr="-init %s",
            extensions=None,
        ),
        interpolation=dict(
            argstr="-%s_final",
        ),
        mean_vol=dict(
            argstr="-meanvol",
        ),
        out_file=dict(
            argstr="-out %s",
            extensions=None,
            genfile=True,
            hash_files=False,
        ),
        output_type=dict(),
        ref_file=dict(
            argstr="-reffile %s",
            extensions=None,
        ),
        ref_vol=dict(
            argstr="-refvol %d",
        ),
        rotation=dict(
            argstr="-rotation %d",
        ),
        save_mats=dict(
            argstr="-mats",
        ),
        save_plots=dict(
            argstr="-plots",
        ),
        save_rms=dict(
            argstr="-rmsabs -rmsrel",
        ),
        scaling=dict(
            argstr="-scaling %.2f",
        ),
        smooth=dict(
            argstr="-smooth %.2f",
        ),
        stages=dict(
            argstr="-stages %d",
        ),
        stats_imgs=dict(
            argstr="-stats",
        ),
        use_contour=dict(
            argstr="-edge",
        ),
        use_gradient=dict(
            argstr="-gdt",
        ),
    )
    inputs = MCFLIRT.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_MCFLIRT_outputs():
    output_map = dict(
        mat_file=dict(),
        mean_img=dict(
            extensions=None,
        ),
        out_file=dict(
            extensions=None,
        ),
        par_file=dict(
            extensions=None,
        ),
        rms_files=dict(),
        std_img=dict(
            extensions=None,
        ),
        variance_img=dict(
            extensions=None,
        ),
    )
    outputs = MCFLIRT.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Algorithms to compute confounds in :abbr:`fMRI (functional MRI)`
"""
import os
import os.path as op
from collections import OrderedDict
from itertools import chain

import nibabel as nb
import numpy as np
from numpy.polynomial import Legendre

from .. import config, logging
from ..external.due import BibTeX
from ..interfaces.base import (
    traits,
    TraitedSpec,
    BaseInterface,
    BaseInterfaceInputSpec,
    File,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
    SimpleInterface,
)
from ..utils.misc import normalize_mc_params

IFLOGGER = logging.getLogger("nipype.interface")


def fallback_svd(a, full_matrices=True, compute_uv=True):
    try:
        return np.linalg.svd(a, full_matrices=full_matrices, compute_uv=compute_uv)
    except np.linalg.LinAlgError:
        pass

    from scipy.linalg import svd

    return svd(
        a, full_matrices=full_matrices, compute_uv=compute_uv, lapack_driver="gesvd"
    )


class ComputeDVARSInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="functional data, after HMC")
    in_mask = File(exists=True, mandatory=True, desc="a brain mask")
    remove_zerovariance = traits.Bool(
        True, usedefault=True, desc="remove voxels with zero variance"
    )
    save_std = traits.Bool(True, usedefault=True, desc="save standardized DVARS")
    save_nstd = traits.Bool(False, usedefault=True, desc="save non-standardized DVARS")
    save_vxstd = traits.Bool(
        False, usedefault=True, desc="save voxel-wise standardized DVARS"
    )
    save_all = traits.Bool(False, usedefault=True, desc="output all DVARS")

    series_tr = traits.Float(desc="repetition time in sec.")
    save_plot = traits.Bool(False, usedefault=True, desc="write DVARS plot")
    figdpi = traits.Int(100, usedefault=True, desc="output dpi for the plot")
    figsize = traits.Tuple(
        traits.Float(11.7),
        traits.Float(2.3),
        usedefault=True,
        desc="output figure size",
    )
    figformat = traits.Enum(
        "png", "pdf", "svg", usedefault=True, desc="output format for figures"
    )
    intensity_normalization = traits.Float(
        1000.0,
        usedefault=True,
        desc="Divide value in each voxel at each timepoint "
        "by the median calculated across all voxels"
        "and timepoints within the mask (if specified)"
        "and then multiply by the value specified by"
        "this parameter. By using the default (1000)"
        "output DVARS will be expressed in "
        "x10 % BOLD units compatible with Power et al."
        "2012. Set this to 0 to disable intensity"
        "normalization altogether.",
    )


class ComputeDVARSOutputSpec(TraitedSpec):
    out_std = File(exists=True, desc="output text file")
    out_nstd = File(exists=True, desc="output text file")
    out_vxstd = File(exists=True, desc="output text file")
    out_all = File(exists=True, desc="output text file")
    avg_std = traits.Float()
    avg_nstd = traits.Float()
    avg_vxstd = traits.Float()
    fig_std = File(exists=True, desc="output DVARS plot")
    fig_nstd = File(exists=True, desc="output DVARS plot")
    fig_vxstd = File(exists=True, desc="output DVARS plot")


class ComputeDVARS(BaseInterface):
    """
    Computes the DVARS.
    """

    input_spec = ComputeDVARSInputSpec
    output_spec = ComputeDVARSOutputSpec
    _references = [
        {
            "entry": BibTeX(
                """\
@techreport{nichols_notes_2013,
    address = {Coventry, UK},
    title = {Notes on {Creating} a {Standardized} {Version} of {DVARS}},
    url = {http://www2.warwick.ac.uk/fac/sci/statistics/staff/academic-\
research/nichols/scripts/fsl/standardizeddvars.pdf},
    urldate = {2016-08-16},
    institution = {University of Warwick},
    author = {Nichols, Thomas},
    year = {2013}
}"""
            ),
            "tags": ["method"],
        },
        {
            "entry": BibTeX(
                """\
@article{power_spurious_2012,
    title = {Spurious but systematic correlations in functional connectivity {MRI} networks \
arise from subject motion},
    volume = {59},
    doi = {10.1016/j.neuroimage.2011.10.018},
    number = {3},
    urldate = {2016-08-16},
    journal = {NeuroImage},
    author = {Power, Jonathan D. and Barnes, Kelly A. and Snyder, Abraham Z. and Schlaggar, \
Bradley L. and Petersen, Steven E.},
    year = {2012},
    pages = {2142--2154},
}
"""
            ),
            "tags": ["method"],
        },
    ]

    def __init__(self, **inputs):
        self._results = {}
        super(ComputeDVARS, self).__init__(**inputs)

    def _gen_fname(self, suffix, ext=None):
        fname, in_ext = op.splitext(op.basename(self.inputs.in_file))

        if in_ext == ".gz":
            fname, in_ext2 = op.splitext(fname)
            in_ext = in_ext2 + in_ext

        if ext is None:
            ext = in_ext

        if ext.startswith("."):
            ext = ext[1:]

        return op.abspath("{}_{}.{}".format(fname, suffix, ext))

    def _run_interface(self, runtime):
        dvars = compute_dvars(
            self.inputs.in_file,
            self.inputs.in_mask,
            remove_zerovariance=self.inputs.remove_zerovariance,
            intensity_normalization=self.inputs.intensity_normalization,
        )

        (
            self._results["avg_std"],
            self._results["avg_nstd"],
            self._results["avg_vxstd"],
        ) = np.mean(dvars, axis=1).astype(float)

        tr = None
        if isdefined(self.inputs.series_tr):
            tr = self.inputs.series_tr

        if self.inputs.save_std:
            out_file = self._gen_fname("dvars_std", ext="tsv")
            np.savetxt(out_file, dvars[0], fmt=b"%0.6f")
            self._results["out_std"] = out_file

            if self.inputs.save_plot:
                self._results["fig_std"] = self._gen_fname(
                    "dvars_std", ext=self.inputs.figformat
                )
                fig = plot_confound(
                    dvars[0], self.inputs.figsize, "Standardized DVARS", series_tr=tr
                )
                fig.savefig(
                    self._results["fig_std"],
                    dpi=float(self.inputs.figdpi),
                    format=self.inputs.figformat,
                    bbox_inches="tight",
                )
                fig.clf()

        if self.inputs.save_nstd:
            out_file = self._gen_fname("dvars_nstd", ext="tsv")
            np.savetxt(out_file, dvars[1], fmt=b"%0.6f")
            self._results["out_nstd"] = out_file

            if self.inputs.save_plot:
                self._results["fig_nstd"] = self._gen_fname(
                    "dvars_nstd", ext=self.inputs.figformat
                )
                fig = plot_confound(
                    dvars[1], self.inputs.figsize, "DVARS", series_tr=tr
                )
                fig.savefig(
                    self._results["fig_nstd"],
                    dpi=float(self.inputs.figdpi),
                    format=self.inputs.figformat,
                    bbox_inches="tight",
                )
                fig.clf()

        if self.inputs.save_vxstd:
            out_file = self._gen_fname("dvars_vxstd", ext="tsv")
            np.savetxt(out_file, dvars[2], fmt=b"%0.6f")
            self._results["out_vxstd"] = out_file

            if self.inputs.save_plot:
                self._results["fig_vxstd"] = self._gen_fname(
                    "dvars_vxstd", ext=self.inputs.figformat
                )
                fig = plot_confound(
                    dvars[2], self.inputs.figsize, "Voxelwise std DVARS", series_tr=tr
                )
                fig.savefig(
                    self._results["fig_vxstd"],
                    dpi=float(self.inputs.figdpi),
                    format=self.inputs.figformat,
                    bbox_inches="tight",
                )
                fig.clf()

        if self.inputs.save_all:
            out_file = self._gen_fname("dvars", ext="tsv")
            np.savetxt(
                out_file,
                np.vstack(dvars).T,
                fmt=b"%0.8f",
                delimiter=b"\t",
                header="std DVARS\tnon-std DVARS\tvx-wise std DVARS",
                comments="",
            )
            self._results["out_all"] = out_file

        return runtime

    def _list_outputs(self):
        return self._results


class FramewiseDisplacementInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="motion parameters")
    parameter_source = traits.Enum(
        "FSL",
        "AFNI",
        "SPM",
        "FSFAST",
        "NIPY",
        desc="Source of movement parameters",
        mandatory=True,
    )
    radius = traits.Float(
        50,
        usedefault=True,
        desc="radius in mm to calculate angular FDs, 50mm is the "
        "default since it is used in Power et al. 2012",
    )
    out_file = File("fd_power_2012.txt", usedefault=True, desc="output file name")
    out_figure = File("fd_power_2012.pdf", usedefault=True, desc="output figure name")
    series_tr = traits.Float(desc="repetition time in sec.")
    save_plot = traits.Bool(False, usedefault=True, desc="write FD plot")
    normalize = traits.Bool(False, usedefault=True, desc="calculate FD in mm/s")
    figdpi = traits.Int(100, usedefault=True, desc="output dpi for the FD plot")
    figsize = traits.Tuple(
        traits.Float(11.7),
        traits.Float(2.3),
        usedefault=True,
        desc="output figure size",
    )


class FramewiseDisplacementOutputSpec(TraitedSpec):
    out_file = File(desc="calculated FD per timestep")
    out_figure = File(desc="output image file")
    fd_average = traits.Float(desc="average FD")


class FramewiseDisplacement(BaseInterface):
    """
    Calculate the :abbr:`FD (framewise displacement)` as in [Power2012]_.
    This implementation reproduces the calculation in fsl_motion_outliers

    .. [Power2012] Power et al., Spurious but systematic correlations in functional
         connectivity MRI networks arise from subject motion, NeuroImage 59(3),
         2012. doi:`10.1016/j.neuroimage.2011.10.018
         <https://doi.org/10.1016/j.neuroimage.2011.10.018>`_.


    """

    input_spec = FramewiseDisplacementInputSpec
    output_spec = FramewiseDisplacementOutputSpec

    _references = [
        {
            "entry": BibTeX(
                """\
@article{power_spurious_2012,
    title = {Spurious but systematic correlations in functional connectivity {MRI} networks \
arise from subject motion},
    volume = {59},
    doi = {10.1016/j.neuroimage.2011.10.018},
    number = {3},
    urldate = {2016-08-16},
    journal = {NeuroImage},
    author = {Power, Jonathan D. and Barnes, Kelly A. and Snyder, Abraham Z. and Schlaggar, \
Bradley L. and Petersen, Steven E.},
    year = {2012},
    pages = {2142--2154},
}
"""
            ),
            "tags": ["method"],
        }
    ]

    def _run_interface(self, runtime):
        mpars = np.loadtxt(self.inputs.in_file)  # mpars is N_t x 6
        mpars = np.apply_along_axis(
            func1d=normalize_mc_params,
            axis=1,
            arr=mpars,
            source=self.inputs.parameter_source,
        )
        diff = mpars[:-1, :6] - mpars[1:, :6]
        diff[:, 3:6] *= self.inputs.radius
        fd_res = np.abs(diff).sum(axis=1)

        self._results = {
            "out_file": op.abspath(self.inputs.out_file),
            "fd_average": float(fd_res.mean()),
        }
        np.savetxt(
            self.inputs.out_file, fd_res, header="FramewiseDisplacement", comments=""
        )

        if self.inputs.save_plot:
            tr = None
            if isdefined(self.inputs.series_tr):
                tr = self.inputs.series_tr

            if self.inputs.normalize and tr is None:
                IFLOGGER.warning("FD plot cannot be normalized if TR is not set")

            self._results["out_figure"] = op.abspath(self.inputs.out_figure)
            fig = plot_confound(
                fd_res,
                self.inputs.figsize,
                "FD",
                units="mm",
                series_tr=tr,
                normalize=self.inputs.normalize,
            )
            fig.savefig(
                self._results["out_figure"],
                dpi=float(self.inputs.figdpi),
                format=self.inputs.out_figure[-3:],
                bbox_inches="tight",
            )
            fig.clf()

        return runtime

    def _list_outputs(self):
        return self._results


class CompCorInputSpec(BaseInterfaceInputSpec):
    realigned_file = File(
        exists=True, mandatory=True, desc="already realigned brain image (4D)"
    )
    mask_files = InputMultiPath(
        File(exists=True),
        desc=(
            "One or more mask files that determines "
            "ROI (3D). When more that one file is "
            "provided ``merge_method`` or "
            "``merge_index`` must be provided"
        ),
    )
    merge_method = traits.Enum(
        "union",
        "intersect",
        "none",
        xor=["mask_index"],
        requires=["mask_files"],
        desc=(
            "Merge method if multiple masks are "
            "present - ``union`` uses voxels included in"
            " at least one input mask, ``intersect`` "
            "uses only voxels present in all input "
            "masks, ``none`` performs CompCor on "
            "each mask individually"
        ),
    )
    mask_index = traits.Range(
        low=0,
        xor=["merge_method"],
        requires=["mask_files"],
        desc="Position of mask in ``mask_files`` to use - first is the default.",
    )
    mask_names = traits.List(
        traits.Str,
        desc="Names for provided masks (for printing into metadata). "
        "If provided, it must be as long as the final mask list "
        "(after any merge and indexing operations).",
    )
    components_file = traits.Str(
        "components_file.txt",
        usedefault=True,
        desc="Filename to store physiological components",
    )
    num_components = traits.Either(
        "all",
        traits.Range(low=1),
        xor=["variance_threshold"],
        desc="Number of components to return from the decomposition. If "
        "``num_components`` is ``all``, then all components will be "
        "retained.",
    )
    # 6 for BOLD, 4 for ASL
    # automatically instantiated to 6 in CompCor below if neither
    # ``num_components`` nor ``variance_threshold`` is defined (for
    # backward compatibility)
    variance_threshold = traits.Range(
        low=0.0,
        high=1.0,
        exclude_low=True,
        exclude_high=True,
        xor=["num_components"],
        desc="Select the number of components to be returned automatically "
        "based on their ability to explain variance in the dataset. "
        "``variance_threshold`` is a fractional value between 0 and 1; "
        "the number of components retained will be equal to the minimum "
        "number of components necessary to explain the provided "
        "fraction of variance in the masked time series.",
    )
    pre_filter = traits.Enum(
        "polynomial",
        "cosine",
        False,
        usedefault=True,
        desc="Detrend time series prior to component " "extraction",
    )
    use_regress_poly = traits.Bool(
        deprecated="0.15.0",
        new_name="pre_filter",
        desc=("use polynomial regression " "pre-component extraction"),
    )
    regress_poly_degree = traits.Range(
        low=1, value=1, usedefault=True, desc="the degree polynomial to use"
    )
    header_prefix = traits.Str(
        desc=(
            "the desired header for the output tsv "
            "file (one column). If undefined, will "
            'default to "CompCor"'
        )
    )
    high_pass_cutoff = traits.Float(
        128, usedefault=True, desc='Cutoff (in seconds) for "cosine" pre-filter'
    )
    repetition_time = traits.Float(
        desc="Repetition time (TR) of series - derived from image header if "
        "unspecified"
    )
    save_pre_filter = traits.Either(
        traits.Bool,
        File,
        default=False,
        usedefault=True,
        desc="Save pre-filter basis as text file",
    )
    save_metadata = traits.Either(
        traits.Bool,
        File,
        default=False,
        usedefault=True,
        desc="Save component metadata as text file",
    )
    ignore_initial_volumes = traits.Range(
        low=0, usedefault=True, desc="Number of volumes at start of series to ignore"
    )
    failure_mode = traits.Enum(
        "error",
        "NaN",
        usedefault=True,
        desc="When no components are found or convergence fails, raise an error "
        "or silently return columns of NaNs.",
    )


class CompCorOutputSpec(TraitedSpec):
    components_file = File(
        exists=True, desc="text file containing the noise components"
    )
    pre_filter_file = File(desc="text file containing high-pass filter basis")
    metadata_file = File(desc="text file containing component metadata")


class CompCor(SimpleInterface):
    """
    Interface with core CompCor computation, used in aCompCor and tCompCor.

    CompCor provides three pre-filter options, all of which include per-voxel
    mean removal:

      - ``'polynomial'``: Legendre polynomial basis
      - ``'cosine'``: Discrete cosine basis
      - ``False``: mean-removal only

    In the case of ``polynomial`` and ``cosine`` filters, a pre-filter file may
    be saved with a row for each volume/timepoint, and a column for each
    non-constant regressor.
    If no non-constant (mean-removal) columns are used, this file may be empty.

    If ``ignore_initial_volumes`` is set, then the specified number of initial
    volumes are excluded both from pre-filtering and CompCor component
    extraction.
    Each column in the components and pre-filter files are prefixe with zeros
    for each excluded volume so that the number of rows continues to match the
    number of volumes in the input file.
    In addition, for each excluded volume, a column is added to the pre-filter
    file with a 1 in the corresponding row.

    Example
    -------
    >>> ccinterface = CompCor()
    >>> ccinterface.inputs.realigned_file = 'functional.nii'
    >>> ccinterface.inputs.mask_files = 'mask.nii'
    >>> ccinterface.inputs.num_components = 1
    >>> ccinterface.inputs.pre_filter = 'polynomial'
    >>> ccinterface.inputs.regress_poly_degree = 2

    """

    input_spec = CompCorInputSpec
    output_spec = CompCorOutputSpec
    _references = [
        {
            "tags": ["method", "implementation"],
            "entry": BibTeX(
                """\
@article{compcor_2007,
    title = {A component based noise correction method (CompCor) for BOLD and perfusion based},
    volume = {37},
    number = {1},
    doi = {10.1016/j.neuroimage.2007.04.042},
    urldate = {2016-08-13},
    journal = {NeuroImage},
    author = {Behzadi, Yashar and Restom, Khaled and Liau, Joy and Liu, Thomas T.},
    year = {2007},
    pages = {90-101}
}"""
            ),
        }
    ]

    def __init__(self, *args, **kwargs):
        """exactly the same as compcor except the header"""
        super(CompCor, self).__init__(*args, **kwargs)
        self._header = "CompCor"

    def _run_interface(self, runtime):
        mask_images = []
        if isdefined(self.inputs.mask_files):
            mask_images = combine_mask_files(
                self.inputs.mask_files, self.inputs.merge_method, self.inputs.mask_index
            )

        if self.inputs.use_regress_poly:
            self.inputs.pre_filter = "polynomial"

        # Degree 0 == remove mean; see compute_noise_components
        degree = (
            self.inputs.regress_poly_degree
            if self.inputs.pre_filter == "polynomial"
            else 0
        )

        imgseries = nb.load(self.inputs.realigned_file)

        if len(imgseries.shape) != 4:
            raise ValueError(
                "{} expected a 4-D nifti file. Input {} has "
                "{} dimensions (shape {})".format(
                    self._header,
                    self.inputs.realigned_file,
                    len(imgseries.shape),
                    imgseries.shape,
                )
            )

        if len(mask_images) == 0:
            img = nb.Nifti1Image(
                np.ones(imgseries.shape[:3], dtype=bool),
                affine=imgseries.affine,
                header=imgseries.header,
            )
            mask_images = [img]

        skip_vols = self.inputs.ignore_initial_volumes
        if skip_vols:
            imgseries = imgseries.__class__(
                imgseries.dataobj[..., skip_vols:], imgseries.affine, imgseries.header
            )

        mask_images = self._process_masks(mask_images, imgseries.dataobj)

        TR = 0
        if self.inputs.pre_filter == "cosine":
            if isdefined(self.inputs.repetition_time):
                TR = self.inputs.repetition_time
            else:
                # Derive TR from NIfTI header, if possible
                try:
                    TR = imgseries.header.get_zooms()[3]
                    if imgseries.header.get_xyzt_units()[1] == "msec":
                        TR /= 1000
                except (AttributeError, IndexError):
                    TR = 0

                if TR == 0:
                    raise ValueError(
                        "{} cannot detect repetition time from image - "
                        "Set the repetition_time input".format(self._header)
                    )

        if isdefined(self.inputs.variance_threshold):
            components_criterion = self.inputs.variance_threshold
        elif isdefined(self.inputs.num_components):
            components_criterion = self.inputs.num_components
        else:
            components_criterion = 6
            IFLOGGER.warning(
                "`num_components` and `variance_threshold` are "
                "not defined. Setting number of components to 6 "
                "for backward compatibility. Please set either "
                "`num_components` or `variance_threshold`, as "
                "this feature may be deprecated in the future."
            )

        components, filter_basis, metadata = compute_noise_components(
            imgseries.get_fdata(dtype=np.float32),
            mask_images,
            components_criterion,
            self.inputs.pre_filter,
            degree,
            self.inputs.high_pass_cutoff,
            TR,
            self.inputs.failure_mode,
            self.inputs.mask_names,
        )

        if skip_vols:
            old_comp = components
            nrows = skip_vols + components.shape[0]
            components = np.zeros((nrows, components.shape[1]), dtype=components.dtype)
            components[skip_vols:] = old_comp

        components_file = os.path.join(os.getcwd(), self.inputs.components_file)
        components_header = self._make_headers(components.shape[1])
        np.savetxt(
            components_file,
            components,
            fmt=b"%.10f",
            delimiter="\t",
            header="\t".join(components_header),
            comments="",
        )
        self._results["components_file"] = os.path.join(
            runtime.cwd, self.inputs.components_file
        )

        save_pre_filter = False
        if self.inputs.pre_filter in ["polynomial", "cosine"]:
            save_pre_filter = self.inputs.save_pre_filter

        if save_pre_filter:
            self._results["pre_filter_file"] = save_pre_filter
            if save_pre_filter is True:
                self._results["pre_filter_file"] = os.path.join(
                    runtime.cwd, "pre_filter.tsv"
                )

            ftype = {"polynomial": "Legendre", "cosine": "Cosine"}[
                self.inputs.pre_filter
            ]
            ncols = filter_basis.shape[1] if filter_basis.size > 0 else 0
            header = ["{}{:02d}".format(ftype, i) for i in range(ncols)]
            if skip_vols:
                old_basis = filter_basis
                # nrows defined above
                filter_basis = np.zeros(
                    (nrows, ncols + skip_vols), dtype=filter_basis.dtype
                )
                if old_basis.size > 0:
                    filter_basis[skip_vols:, :ncols] = old_basis
                filter_basis[:skip_vols, -skip_vols:] = np.eye(skip_vols)
                header.extend(
                    ["NonSteadyStateOutlier{:02d}".format(i) for i in range(skip_vols)]
                )
            np.savetxt(
                self._results["pre_filter_file"],
                filter_basis,
                fmt=b"%.10f",
                delimiter="\t",
                header="\t".join(header),
                comments="",
            )

        metadata_file = self.inputs.save_metadata
        if metadata_file:
            self._results["metadata_file"] = metadata_file
            if metadata_file is True:
                self._results["metadata_file"] = os.path.join(
                    runtime.cwd, "component_metadata.tsv"
                )
            components_names = np.empty(len(metadata["mask"]), dtype="object_")
            retained = np.where(metadata["retained"])
            not_retained = np.where(np.logical_not(metadata["retained"]))
            components_names[retained] = components_header
            components_names[not_retained] = [
                "dropped{}".format(i) for i in range(len(not_retained[0]))
            ]
            with open(self._results["metadata_file"], "w") as f:
                f.write("\t".join(["component"] + list(metadata.keys())) + "\n")
                for i in zip(components_names, *metadata.values()):
                    f.write(
                        "{0[0]}\t{0[1]}\t{0[2]:.10f}\t"
                        "{0[3]:.10f}\t{0[4]:.10f}\t{0[5]}\n".format(i)
                    )

        return runtime

    def _process_masks(self, mask_images, timeseries=None):
        return mask_images

    def _make_headers(self, num_col):
        header = (
            self.inputs.header_prefix
            if isdefined(self.inputs.header_prefix)
            else self._header
        )
        headers = ["{}{:02d}".format(header, i) for i in range(num_col)]
        return headers


class ACompCor(CompCor):
    """
    Anatomical compcor: for inputs and outputs, see CompCor.
    When the mask provided is an anatomical mask, then CompCor
    is equivalent to ACompCor.
    """

    def __init__(self, *args, **kwargs):
        """exactly the same as compcor except the header"""
        super(ACompCor, self).__init__(*args, **kwargs)
        self._header = "aCompCor"


class TCompCorInputSpec(CompCorInputSpec):
    # and all the fields in CompCorInputSpec
    percentile_threshold = traits.Range(
        low=0.0,
        high=1.0,
        value=0.02,
        exclude_low=True,
        exclude_high=True,
        usedefault=True,
        desc="the percentile "
        "used to select highest-variance "
        "voxels, represented by a number "
        "between 0 and 1, exclusive. By "
        "default, this value is set to .02. "
        "That is, the 2% of voxels "
        "with the highest variance are used.",
    )


class TCompCorOutputSpec(CompCorOutputSpec):
    # and all the fields in CompCorOutputSpec
    high_variance_masks = OutputMultiPath(
        File(exists=True), desc=(("voxels exceeding the variance" " threshold"))
    )


class TCompCor(CompCor):
    """
    Interface for tCompCor. Computes a ROI mask based on variance of voxels.

    Example
    -------
    >>> ccinterface = TCompCor()
    >>> ccinterface.inputs.realigned_file = 'functional.nii'
    >>> ccinterface.inputs.mask_files = 'mask.nii'
    >>> ccinterface.inputs.num_components = 1
    >>> ccinterface.inputs.pre_filter = 'polynomial'
    >>> ccinterface.inputs.regress_poly_degree = 2
    >>> ccinterface.inputs.percentile_threshold = .03

    """

    input_spec = TCompCorInputSpec
    output_spec = TCompCorOutputSpec

    def __init__(self, *args, **kwargs):
        """exactly the same as compcor except the header"""
        super(TCompCor, self).__init__(*args, **kwargs)
        self._header = "tCompCor"
        self._mask_files = []

    def _process_masks(self, mask_images, timeseries=None):
        out_images = []
        self._mask_files = []
        timeseries = np.asanyarray(timeseries)
        for i, img in enumerate(mask_images):
            mask = np.asanyarray(img.dataobj).astype(bool)
            imgseries = timeseries[mask, :]
            imgseries = regress_poly(2, imgseries)[0]
            tSTD = _compute_tSTD(imgseries, 0, axis=-1)
            threshold_std = np.percentile(
                tSTD,
                np.round(100.0 * (1.0 - self.inputs.percentile_threshold)).astype(int),
            )
            mask_data = np.zeros_like(mask)
            mask_data[mask != 0] = tSTD >= threshold_std
            out_image = nb.Nifti1Image(mask_data, affine=img.affine, header=img.header)

            # save mask
            mask_file = os.path.abspath("mask_{:03d}.nii.gz".format(i))
            out_image.to_filename(mask_file)
            IFLOGGER.debug(
                "tCompcor computed and saved mask of shape %s to " "mask_file %s",
                str(mask.shape),
                mask_file,
            )
            self._mask_files.append(mask_file)
            out_images.append(out_image)
        return out_images

    def _list_outputs(self):
        outputs = super(TCompCor, self)._list_outputs()
        outputs["high_variance_masks"] = self._mask_files
        return outputs


class TSNRInputSpec(BaseInterfaceInputSpec):
    in_file = InputMultiPath(
        File(exists=True),
        mandatory=True,
        desc="realigned 4D file or a list of 3D files",
    )
    regress_poly = traits.Range(low=1, desc="Remove polynomials")
    tsnr_file = File(
        "tsnr.nii.gz", usedefault=True, hash_files=False, desc="output tSNR file"
    )
    mean_file = File(
        "mean.nii.gz", usedefault=True, hash_files=False, desc="output mean file"
    )
    stddev_file = File(
        "stdev.nii.gz", usedefault=True, hash_files=False, desc="output tSNR file"
    )
    detrended_file = File(
        "detrend.nii.gz",
        usedefault=True,
        hash_files=False,
        desc="input file after detrending",
    )


class TSNROutputSpec(TraitedSpec):
    tsnr_file = File(exists=True, desc="tsnr image file")
    mean_file = File(exists=True, desc="mean image file")
    stddev_file = File(exists=True, desc="std dev image file")
    detrended_file = File(desc="detrended input file")


class TSNR(BaseInterface):
    """
    Computes the time-course SNR for a time series

    Typically you want to run this on a realigned time-series.

    Example
    -------
    >>> tsnr = TSNR()
    >>> tsnr.inputs.in_file = 'functional.nii'
    >>> res = tsnr.run() # doctest: +SKIP

    """

    input_spec = TSNRInputSpec
    output_spec = TSNROutputSpec

    def _run_interface(self, runtime):
        img = nb.load(self.inputs.in_file[0])
        header = img.header.copy()
        vollist = [nb.load(filename) for filename in self.inputs.in_file]
        data = np.concatenate(
            [
                vol.get_fdata(dtype=np.float32).reshape(vol.shape[:3] + (-1,))
                for vol in vollist
            ],
            axis=3,
        )
        data = np.nan_to_num(data)

        if data.dtype.kind == "i":
            header.set_data_dtype(np.float32)
            data = data.astype(np.float32)

        if isdefined(self.inputs.regress_poly):
            data = regress_poly(self.inputs.regress_poly, data, remove_mean=False)[0]
            img = nb.Nifti1Image(data, img.affine, header)
            nb.save(img, op.abspath(self.inputs.detrended_file))

        meanimg = np.mean(data, axis=3)
        stddevimg = np.std(data, axis=3)
        tsnr = np.zeros_like(meanimg)
        stddevimg_nonzero = stddevimg > 1.0e-3
        tsnr[stddevimg_nonzero] = (
            meanimg[stddevimg_nonzero] / stddevimg[stddevimg_nonzero]
        )
        img = nb.Nifti1Image(tsnr, img.affine, header)
        nb.save(img, op.abspath(self.inputs.tsnr_file))
        img = nb.Nifti1Image(meanimg, img.affine, header)
        nb.save(img, op.abspath(self.inputs.mean_file))
        img = nb.Nifti1Image(stddevimg, img.affine, header)
        nb.save(img, op.abspath(self.inputs.stddev_file))
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        for k in ["tsnr_file", "mean_file", "stddev_file"]:
            outputs[k] = op.abspath(getattr(self.inputs, k))

        if isdefined(self.inputs.regress_poly):
            outputs["detrended_file"] = op.abspath(self.inputs.detrended_file)
        return outputs


class NonSteadyStateDetectorInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="4D NIFTI EPI file")


class NonSteadyStateDetectorOutputSpec(TraitedSpec):
    n_volumes_to_discard = traits.Int(
        desc="Number of non-steady state volumes"
        "detected in the beginning of the scan."
    )


class NonSteadyStateDetector(BaseInterface):
    """
    Returns the number of non-steady state volumes detected at the beginning
    of the scan.
    """

    input_spec = NonSteadyStateDetectorInputSpec
    output_spec = NonSteadyStateDetectorOutputSpec

    def _run_interface(self, runtime):
        in_nii = nb.load(self.inputs.in_file)
        global_signal = (
            in_nii.dataobj[:, :, :, :50].mean(axis=0).mean(axis=0).mean(axis=0)
        )

        self._results = {"n_volumes_to_discard": is_outlier(global_signal)}

        return runtime

    def _list_outputs(self):
        return self._results


def compute_dvars(
    in_file, in_mask, remove_zerovariance=False, intensity_normalization=1000
):
    """
    Compute the :abbr:`DVARS (D referring to temporal
    derivative of timecourses, VARS referring to RMS variance over voxels)`
    [Power2012]_.

    Particularly, the *standardized* :abbr:`DVARS (D referring to temporal
    derivative of timecourses, VARS referring to RMS variance over voxels)`
    [Nichols2013]_ are computed.

    .. [Nichols2013] Nichols T, `Notes on creating a standardized version of
         DVARS <http://www2.warwick.ac.uk/fac/sci/statistics/staff/academic-\
research/nichols/scripts/fsl/standardizeddvars.pdf>`_, 2013.

    .. note:: Implementation details

      Uses the implementation of the `Yule-Walker equations
      from nitime
      <http://nipy.org/nitime/api/generated/nitime.algorithms.autoregressive.html\
#nitime.algorithms.autoregressive.AR_est_YW>`_
      for the :abbr:`AR (auto-regressive)` filtering of the fMRI signal.

    :param numpy.ndarray func: functional data, after head-motion-correction.
    :param numpy.ndarray mask: a 3D mask of the brain
    :param bool output_all: write out all dvars
    :param str out_file: a path to which the standardized dvars should be saved.
    :return: the standardized DVARS

    """
    import numpy as np
    import nibabel as nb
    from nitime.algorithms import AR_est_YW
    import warnings

    func = nb.load(in_file).get_fdata(dtype=np.float32)
    mask = np.asanyarray(nb.load(in_mask).dataobj).astype(np.uint8)

    if len(func.shape) != 4:
        raise RuntimeError("Input fMRI dataset should be 4-dimensional")

    idx = np.where(mask > 0)
    mfunc = func[idx[0], idx[1], idx[2], :]

    if intensity_normalization != 0:
        mfunc = (mfunc / np.median(mfunc)) * intensity_normalization

    # Robust standard deviation (we are using "lower" interpolation
    # because this is what FSL is doing
    func_sd = (
        np.percentile(mfunc, 75, axis=1, interpolation="lower")
        - np.percentile(mfunc, 25, axis=1, interpolation="lower")
    ) / 1.349

    if remove_zerovariance:
        mfunc = mfunc[func_sd != 0, :]
        func_sd = func_sd[func_sd != 0]

    # Compute (non-robust) estimate of lag-1 autocorrelation
    ar1 = np.apply_along_axis(
        AR_est_YW, 1, regress_poly(0, mfunc, remove_mean=True)[0].astype(np.float32), 1
    )[:, 0]

    # Compute (predicted) standard deviation of temporal difference time series
    diff_sdhat = np.squeeze(np.sqrt(((1 - ar1) * 2).tolist())) * func_sd
    diff_sd_mean = diff_sdhat.mean()

    # Compute temporal difference time series
    func_diff = np.diff(mfunc, axis=1)

    # DVARS (no standardization)
    dvars_nstd = np.sqrt(np.square(func_diff).mean(axis=0))

    # standardization
    dvars_stdz = dvars_nstd / diff_sd_mean

    with warnings.catch_warnings():  # catch, e.g., divide by zero errors
        warnings.filterwarnings("error")

        # voxelwise standardization
        diff_vx_stdz = np.square(
            func_diff / np.array([diff_sdhat] * func_diff.shape[-1]).T
        )
        dvars_vx_stdz = np.sqrt(diff_vx_stdz.mean(axis=0))

    return (dvars_stdz, dvars_nstd, dvars_vx_stdz)


def plot_confound(tseries, figsize, name, units=None, series_tr=None, normalize=False):
    """
    A helper function to plot :abbr:`fMRI (functional MRI)` confounds.

    """
    import matplotlib

    matplotlib.use(config.get("execution", "matplotlib_backend"))
    import matplotlib.pyplot as plt
    from matplotlib.gridspec import GridSpec
    from matplotlib.backends.backend_pdf import FigureCanvasPdf as FigureCanvas
    import seaborn as sns

    fig = plt.Figure(figsize=figsize)
    FigureCanvas(fig)
    grid = GridSpec(1, 2, width_ratios=[3, 1], wspace=0.025)
    grid.update(hspace=1.0, right=0.95, left=0.1, bottom=0.2)

    ax = fig.add_subplot(grid[0, :-1])
    if normalize and series_tr is not None:
        tseries /= series_tr

    ax.plot(tseries)
    ax.set_xlim((0, len(tseries)))
    ylabel = name
    if units is not None:
        ylabel += (" speed [{}/s]" if normalize else " [{}]").format(units)
    ax.set_ylabel(ylabel)

    xlabel = "Frame #"
    if series_tr is not None:
        xlabel = "Frame # ({} sec TR)".format(series_tr)
    ax.set_xlabel(xlabel)
    ylim = ax.get_ylim()

    ax = fig.add_subplot(grid[0, -1])
    sns.distplot(tseries, vertical=True, ax=ax)
    ax.set_xlabel("Frames")
    ax.set_ylim(ylim)
    ax.set_yticklabels([])
    return fig


def is_outlier(points, thresh=3.5):
    """
    Returns a boolean array with True if points are outliers and False
    otherwise.

    :param nparray points: an numobservations by numdimensions numpy array of observations
    :param float thresh: the modified z-score to use as a threshold. Observations with
        a modified z-score (based on the median absolute deviation) greater
        than this value will be classified as outliers.

    :return: A bolean mask, of size numobservations-length array.

    .. note:: References

        Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and
        Handle Outliers", The ASQC Basic References in Quality Control:
        Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.

    """
    if len(points.shape) == 1:
        points = points[:, None]
    median = np.median(points, axis=0)
    diff = np.sum((points - median) ** 2, axis=-1)
    diff = np.sqrt(diff)
    med_abs_deviation = np.median(diff)

    modified_z_score = 0.6745 * diff / med_abs_deviation

    timepoints_to_discard = 0
    for i in range(len(modified_z_score)):
        if modified_z_score[i] <= thresh:
            break
        else:
            timepoints_to_discard += 1

    return timepoints_to_discard


def cosine_filter(
    data, timestep, period_cut, remove_mean=True, axis=-1, failure_mode="error"
):
    datashape = data.shape
    timepoints = datashape[axis]
    if datashape[0] == 0 and failure_mode != "error":
        return data, np.array([])

    data = data.reshape((-1, timepoints))

    frametimes = timestep * np.arange(timepoints)
    X = _full_rank(_cosine_drift(period_cut, frametimes))[0]
    non_constant_regressors = X[:, :-1] if X.shape[1] > 1 else np.array([])

    betas = np.linalg.lstsq(X, data.T)[0]

    if not remove_mean:
        X = X[:, :-1]
        betas = betas[:-1]

    residuals = data - X.dot(betas).T

    return residuals.reshape(datashape), non_constant_regressors


def regress_poly(degree, data, remove_mean=True, axis=-1, failure_mode="error"):
    """
    Returns data with degree polynomial regressed out.

    :param bool remove_mean: whether or not demean data (i.e. degree 0),
    :param int axis: numpy array axes along which regression is performed

    """
    IFLOGGER.debug(
        "Performing polynomial regression on data of shape %s", str(data.shape)
    )

    datashape = data.shape
    timepoints = datashape[axis]
    if datashape[0] == 0 and failure_mode != "error":
        return data, np.array([])

    # Rearrange all voxel-wise time-series in rows
    data = data.reshape((-1, timepoints))

    # Generate design matrix
    X = np.ones((timepoints, 1))  # quick way to calc degree 0
    for i in range(degree):
        polynomial_func = Legendre.basis(i + 1)
        value_array = np.linspace(-1, 1, timepoints)
        X = np.hstack((X, polynomial_func(value_array)[:, np.newaxis]))

    non_constant_regressors = X[:, :-1] if X.shape[1] > 1 else np.array([])

    # Calculate coefficients
    betas = np.linalg.pinv(X).dot(data.T)

    # Estimation
    if remove_mean:
        datahat = X.dot(betas).T
    else:  # disregard the first layer of X, which is degree 0
        datahat = X[:, 1:].dot(betas[1:, ...]).T
    regressed_data = data - datahat

    # Back to original shape
    return regressed_data.reshape(datashape), non_constant_regressors


def combine_mask_files(mask_files, mask_method=None, mask_index=None):
    """Combines input mask files into a single nibabel image

    A helper function for CompCor

    Parameters
    ----------
    mask_files: a list
        one or more binary mask files
    mask_method: enum ('union', 'intersect', 'none')
        determines how to combine masks
    mask_index: an integer
        determines which file to return (mutually exclusive with mask_method)

    Returns
    -------
    masks: a list of nibabel images

    """

    if isdefined(mask_index) or not isdefined(mask_method):
        if not isdefined(mask_index):
            if len(mask_files) == 1:
                mask_index = 0
            else:
                raise ValueError(
                    (
                        "When more than one mask file is provided, "
                        "one of merge_method or mask_index must be "
                        "set"
                    )
                )
        if mask_index < len(mask_files):
            mask = nb.load(mask_files[mask_index])
            return [mask]
        raise ValueError(
            ("mask_index {0} must be less than number of mask " "files {1}").format(
                mask_index, len(mask_files)
            )
        )
    masks = []
    if mask_method == "none":
        for filename in mask_files:
            masks.append(nb.load(filename))
        return masks

    if mask_method == "union":
        mask = None
        for filename in mask_files:
            img = nb.load(filename)
            img_as_mask = np.asanyarray(img.dataobj).astype("int32") > 0
            if mask is None:
                mask = img_as_mask
            np.logical_or(mask, img_as_mask, mask)
        img = nb.Nifti1Image(mask, img.affine, header=img.header)
        return [img]

    if mask_method == "intersect":
        mask = None
        for filename in mask_files:
            img = nb.load(filename)
            img_as_mask = np.asanyarray(img.dataobj).astype("int32") > 0
            if mask is None:
                mask = img_as_mask
            np.logical_and(mask, img_as_mask, mask)
        img = nb.Nifti1Image(mask, img.affine, header=img.header)
        return [img]


def compute_noise_components(
    imgseries,
    mask_images,
    components_criterion=0.5,
    filter_type=False,
    degree=0,
    period_cut=128,
    repetition_time=None,
    failure_mode="error",
    mask_names=None,
):
    """
    Compute the noise components from the image series for each mask.

    Parameters
    ----------
    imgseries: nibabel image
        Time series data to be decomposed.
    mask_images: list
        List of nibabel images. Time series data from ``img_series`` is subset
        according to the spatial extent of each mask, and the subset data is
        then decomposed using principal component analysis. Masks should be
        coextensive with either anatomical or spatial noise ROIs.
    components_criterion: float
        Number of noise components to return. If this is a decimal value
        between 0 and 1, then ``create_noise_components`` will instead return
        the smallest number of components necessary to explain the indicated
        fraction of variance. If ``components_criterion`` is ``all``, then all
        components will be returned.
    filter_type: str
        Type of filter to apply to time series before computing noise components.

            - 'polynomial' - Legendre polynomial basis
            - 'cosine' - Discrete cosine (DCT) basis
            - False - None (mean-removal only)

    failure_mode: str
        Action to be taken in the event that any decomposition fails to
        identify any components. ``error`` indicates that the routine should
        raise an exception and exit, while any other value indicates that the
        routine should return a matrix of NaN values equal in size to the
        requested decomposition matrix.
    mask_names: list or None
        List of names for each image in ``mask_images``. This should be equal in
        length to ``mask_images``, with the ith element of ``mask_names`` naming
        the ith element of ``mask_images``.
    degree: int
        Order of polynomial used to remove trends from the timeseries
    period_cut: float
        Minimum period (in sec) for DCT high-pass filter
    repetition_time: float
        Time (in sec) between volume acquisitions. This must be defined if
        the ``filter_type`` is ``cosine``.

    Returns
    -------
    components: numpy array
        Numpy array containing the requested set of noise components
    basis: numpy array
        Numpy array containing the (non-constant) filter regressors
    metadata: OrderedDict{str: numpy array}
        Dictionary of eigenvalues, fractional explained variances, and
        cumulative explained variances.

    """
    basis = np.array([])
    if components_criterion == "all":
        components_criterion = -1
    mask_names = mask_names or range(len(mask_images))

    components = []
    md_mask = []
    md_sv = []
    md_var = []
    md_cumvar = []
    md_retained = []

    for name, img in zip(mask_names, mask_images):
        mask = np.asanyarray(nb.squeeze_image(img).dataobj).astype(bool)
        if imgseries.shape[:3] != mask.shape:
            raise ValueError(
                "Inputs for CompCor, timeseries and mask, do not have "
                "matching spatial dimensions ({} and {}, respectively)".format(
                    imgseries.shape[:3], mask.shape
                )
            )

        voxel_timecourses = imgseries[mask, :]

        # Zero-out any bad values
        voxel_timecourses[np.isnan(np.sum(voxel_timecourses, axis=1)), :] = 0

        # Currently support Legendre-polynomial or cosine or detrending
        # With no filter, the mean is nonetheless removed (poly w/ degree 0)
        if filter_type == "cosine":
            if repetition_time is None:
                raise ValueError("Repetition time must be provided for cosine filter")
            voxel_timecourses, basis = cosine_filter(
                voxel_timecourses,
                repetition_time,
                period_cut,
                failure_mode=failure_mode,
            )
        elif filter_type in ("polynomial", False):
            # from paper:
            # "The constant and linear trends of the columns in the matrix M were
            # removed [prior to ...]"
            voxel_timecourses, basis = regress_poly(
                degree, voxel_timecourses, failure_mode=failure_mode
            )

        # "Voxel time series from the noise ROI (either anatomical or tSTD) were
        # placed in a matrix M of size Nxm, with time along the row dimension
        # and voxels along the column dimension."
        M = voxel_timecourses.T

        # "[... were removed] prior to column-wise variance normalization."
        M = M / _compute_tSTD(M, 1.0)

        # "The covariance matrix C = MMT was constructed and decomposed into its
        # principal components using a singular value decomposition."
        try:
            u, s, _ = fallback_svd(M, full_matrices=False)
        except (np.linalg.LinAlgError, ValueError):
            if failure_mode == "error":
                raise
            s = np.full(M.shape[0], np.nan, dtype=np.float32)
            if components_criterion >= 1:
                u = np.full(
                    (M.shape[0], components_criterion), np.nan, dtype=np.float32
                )
            else:
                u = np.full((M.shape[0], 1), np.nan, dtype=np.float32)

        variance_explained = (s**2) / np.sum(s**2)
        cumulative_variance_explained = np.cumsum(variance_explained)

        num_components = int(components_criterion)
        if 0 < components_criterion < 1:
            num_components = (
                np.searchsorted(cumulative_variance_explained, components_criterion) + 1
            )
        elif components_criterion == -1:
            num_components = len(s)

        num_components = int(num_components)
        if num_components == 0:
            break

        components.append(u[:, :num_components])
        md_mask.append([name] * len(s))
        md_sv.append(s)
        md_var.append(variance_explained)
        md_cumvar.append(cumulative_variance_explained)
        md_retained.append([i < num_components for i in range(len(s))])

    if len(components) > 0:
        components = np.hstack(components)
    else:
        if failure_mode == "error":
            raise ValueError("No components found")
        components = np.full((M.shape[0], num_components), np.nan, dtype=np.float32)

    metadata = OrderedDict(
        [
            ("mask", list(chain(*md_mask))),
            ("singular_value", np.hstack(md_sv)),
            ("variance_explained", np.hstack(md_var)),
            ("cumulative_variance_explained", np.hstack(md_cumvar)),
            ("retained", list(chain(*md_retained))),
        ]
    )

    return components, basis, metadata


def _compute_tSTD(M, x, axis=0):
    stdM = np.std(M, axis=axis)
    # set bad values to x
    stdM[stdM == 0] = x
    stdM[np.isnan(stdM)] = x
    return stdM


# _cosine_drift and _full_rank copied from nipy/modalities/fmri/design_matrix
#
# Nipy release: 0.4.1
# Modified for smooth integration in CompCor classes


def _cosine_drift(period_cut, frametimes):
    """Create a cosine drift matrix with periods greater or equals to period_cut

    Parameters
    ----------
    period_cut: float
         Cut period of the low-pass filter (in sec)
    frametimes: array of shape(nscans)
         The sampling times (in sec)

    Returns
    -------
    cdrift:  array of shape(n_scans, n_drifts)
             cosin drifts plus a constant regressor at cdrift[:,0]

    Ref: http://en.wikipedia.org/wiki/Discrete_cosine_transform DCT-II
    """
    len_tim = len(frametimes)
    n_times = np.arange(len_tim)
    hfcut = 1.0 / period_cut  # input parameter is the period

    # frametimes.max() should be (len_tim-1)*dt
    dt = frametimes[1] - frametimes[0]
    # hfcut = 1/(2*dt) yields len_time
    # If series is too short, return constant regressor
    order = max(int(np.floor(2 * len_tim * hfcut * dt)), 1)
    cdrift = np.zeros((len_tim, order))
    nfct = np.sqrt(2.0 / len_tim)

    for k in range(1, order):
        cdrift[:, k - 1] = nfct * np.cos((np.pi / len_tim) * (n_times + 0.5) * k)

    cdrift[:, order - 1] = 1.0  # or 1./sqrt(len_tim) to normalize
    return cdrift


def _full_rank(X, cmax=1e15):
    """
    This function possibly adds a scalar matrix to X
    to guarantee that the condition number is smaller than a given threshold.

    Parameters
    ----------
    X: array of shape(nrows, ncols)
    cmax=1.e-15, float tolerance for condition number

    Returns
    -------
    X: array of shape(nrows, ncols) after regularization
    cmax=1.e-15, float tolerance for condition number
    """
    U, s, V = fallback_svd(X, full_matrices=False)
    smax, smin = s.max(), s.min()
    c = smax / smin
    if c < cmax:
        return X, c
    IFLOGGER.warning("Matrix is singular at working precision, regularizing...")
    lda = (smax - cmax * smin) / (cmax - 1)
    s = s + lda
    X = np.dot(U, np.dot(np.diag(s), V))
    return X, cmax
